r"""
ZK proof for linear representations of discrete logarithms, our basic building block.

An example of such proof is :math:`PK\{ (x_0, x_1): y = x_0 G_0 + x_1 G_1 \}`, where :math:`x_0` and
:math:`x_1` are secret integers from a finite field, :math:`G_0` and :math:`G_1` are points on a
same elliptic curve, and :math:`y` is the actual value of the expression :math:`x_0 G_0 + x_1 G_1`.

See "`Proof Systems for General Statements about Discrete Logarithms`_" by Camenisch and Stadler,
1997 for the details.

.. _`Proof Systems for General Statements about Discrete Logarithms`:
    ftp://ftp.inf.ethz.ch/pub/crypto/publications/CamSta97b.pdf

"""
from hashlib import sha256

from petlib.bn import Bn

from zksk.base import Verifier, Prover, SimulationTranscript
from zksk.expr import Secret, Expression
from zksk.utils import get_random_num
from zksk.consts import CHALLENGE_LENGTH
from zksk.composition import ComposableProofStmt
from zksk.exceptions import IncompleteValuesError, InvalidExpression

import warnings


class DLRepVerifier(Verifier):
    def check_responses_consistency(self, responses, responses_dict=None):
        """
        Check if reoccuring secrets yield the same responses.

        To do so, go through the names of the secrets in the current DLRep, and construct a mapping
        between secrets and responses.

        Args:
            response: List of responses
            responses_dict: Mapping from secrets to responses

        Returns:
            bool: True if responses are consistent, False otherwise.
        """
        if responses_dict is None:
            responses_dict = {}

        for i, s in enumerate(self.stmt.secret_vars):
            if s in responses_dict.keys():
                if responses[i] != responses_dict[s]:
                    return False
            else:
                responses_dict.update({s: responses[i]})
        return True


class DLRep(ComposableProofStmt):
    """
    Proof statement for a discrete-logarithm representation proof.

    Supports statements of the following form:

    .. math::
        PK\{ (x_0, x_1, ..., x_n): Y = x_0 G_0 + x_1 G_1 + ... + x_n G_n \}

    Example usage for :math:`PK\{x: Y = x G \}`:

    >>> from petlib.ec import EcGroup
    >>> x = Secret(name="x")
    >>> g = EcGroup().generator()
    >>> y = 42 * g
    >>> stmt = DLRep(y, x * g)
    >>> nizk = stmt.prove({x: 42})
    >>> stmt.verify(nizk)
    True

    Args:
        expr (:py:class:`zksk.base.Expression`): Proof statement.
            For example: ``Secret("x") * g`` represents :math:`PK\{ x: Y = x G \}`.
        lhs: "Left-hand side." Value of :math:`Y`.
    """

    verifier_cls = DLRepVerifier

    def __init__(self, lhs, expr, simulated=False):
        if isinstance(expr, Expression):
            self.bases = list(expr.bases)
            self.secret_vars = list(expr.secrets)
        else:
            raise TypeError("Expected an Expression. Got: {}".format(expr))

        # Check all the generators live in the same group
        test_group = self.bases[0].group
        for g in self.bases:
            if g.group != test_group:
                raise InvalidExpression(
                    "All bases should come from the same group", g.group
                )

        # Construct a dictionary with the secret values we already know
        self.secret_values = {}
        for sec in self.secret_vars:
            if sec.value is not None:
                self.secret_values[sec] = sec.value

        self.lhs = lhs
        self.set_simulated(simulated)

    def get_prover(self, secrets_dict=None):
        """
        Get a prover for the current proof statement.

        Args:
            secrets_dict: Optional mapping from secrets or secret names to their values.

        Returns:
            :py:class:`DLRepProver` or None: Prover object if all secret values are known.
        """
        if secrets_dict is None:
            secrets_dict = {}

        # First we update the dictionary we have with the additional secrets, and process it
        self.secret_values.update(secrets_dict)
        secrets_dict = self.secret_values
        # If missing secrets or simulation parameter set, return now
        if (
            self.set_simulated()
            or secrets_dict == {}
            or any(sec not in secrets_dict.keys() for sec in set(self.secret_vars))
        ):
            # TODO: Make this raise:
            # raise IncompleteValuesError(self.secret_vars)
            return None

        # We check everything is indeed a big number, else we cast it
        for name, sec in secrets_dict.items():
            if not isinstance(sec, Bn):
                secrets_dict[name] = Bn(sec)

        return DLRepProver(self, secrets_dict)

    def get_proof_id(self, secret_id_map=None):
        """
        Identifier for the proof statement

        Returns:
            list: Objects that can be used for hashing.
        """
        proof_id = super().get_proof_id(secret_id_map)
        return proof_id + [self.lhs]

    def get_randomizers(self):
        """
        Initialize randomizers for each secret.

        Each randomizer is drawn at random from the associated group.

        By using a dictionary, we enforce that if secret are repeated in :math:`x_0 G_0 + x_1 G_1 +
        ... + x_n G_n`, that is, if :math:`x_i` and :math:`x_j` have the same name, they will get
        the same random value. Identical secret values and identical randomizers will yield
        identical responses, and this identity will be checked by the verifier.

        Returns:
            dict: Mapping from secrets to the random values that are needed to compute the responses
                of the proof.
        """
        output = {}
        order = self.bases[0].group.order()
        for sec in set(self.secret_vars):
            output.update({sec: order.random()})
        return output

    def recompute_commitment(self, challenge, responses):

        commitment = (
            self.lhs.group.wsum(responses, self.bases) + (-challenge) * self.lhs
        )
        return commitment

    def simulate_proof(self, responses_dict=None, challenge=None):
        """
        Returns a transcript of a proof simulation. Responses and challenge can be enforced.  The
        function will misbehave if passed a non-empty but incomplete responses_dict.

        Args:
            responses_dict: Optinal mapping from secrets or secret names to responses.
            challenge: Optional challenge to use in the simulation
        """
        # Fill the missing positions of the responses dictionary
        responses_dict = self.update_randomizers(responses_dict)
        if challenge is None:
            challenge = get_random_num(CHALLENGE_LENGTH)

        responses = [responses_dict[m] for m in self.secret_vars]
        # Random responses, the same for shared secrets
        commitment = self.recompute_commitment(challenge, responses)

        return SimulationTranscript(
            commitment=commitment, challenge=challenge, responses=responses
        )


class DLRepProver(Prover):
    """The prover in a discrete logarithm proof."""

    def internal_commit(self, randomizers_dict=None):
        """
        Compute the commitment using the randomizers.

        Args:
            randomizers_dict: Optional mapping from secrets or secret names to random values. Every
                random value not given here will be generated at random.

        Returns:
            A single commitment---sum of bases, weighted by the corresponding randomizers
        """
        # Fill the missing positions of the randomizers dictionary
        randomizers_dict = self.stmt.update_randomizers(randomizers_dict)

        # Compute an ordered list of randomizers mirroring the Secret objects
        self.ks = [randomizers_dict[sec] for sec in self.stmt.secret_vars]
        subcommits = [a * b for a, b in zip(self.ks, self.stmt.bases)]

        # We build the commitment doing the product k0 * g0 + k1 * g1...
        result = self.stmt.bases[0].group.infinite()
        for com in subcommits:
            result = result + com

        return result

    def compute_response(self, challenge):
        """
        Constructs an (ordered) list of response for each secret.

        For each secret :math:`x` and a random value :math:`k` (associated to :math:`x`), the
        response is equal to :math:`k + c x`, where :math:`c` is the challenge value.

        Args:
            challenge: Challenge value

        Returns:
            A list of responses
        """
        order = self.stmt.bases[0].group.order()
        resps = [
            (self.secret_values[self.stmt.secret_vars[i]] * challenge + k) % order
            for i, k in enumerate(self.ks)
        ]
        return resps



"""
Utils that can be useful for debugging.
"""


class SigmaProtocol:
    """
    Sigma-protocol runner.

    Args:
        verifier: Verifier object
        prover: Prover object
    """

    def __init__(self, verifier, prover):
        self.verifier = verifier
        self.prover = prover

    def verify(self, verbose=True):
        """Run the verification process."""

        # Funky names.
        victor = self.verifier
        peggy = self.prover

        precommitment = peggy.precommit()
        victor.process_precommitment(precommitment)
        commitment = peggy.commit()
        challenge = victor.send_challenge(commitment)
        response = peggy.compute_response(challenge)
        result = victor.verify(response)

        if verbose:
            if result:
                print("Verified for {0}".format(victor.__class__.__name__))
            else:
                print("Not verified for {0}".format(victor.__class__.__name__))

        return result



"""
ZK proof of knowledge of a BBS+ signature.

This proof can be used to build blacklistable anonymous credential schemes.

See "`Constant-Size Dynamic k-TAA`_" by Au et al., 2008 for the details.

.. _`Constant-Size Dynamic k-TAA`:
   https://eprint.iacr.org/2008/136.pdf

"""

import attr

from zksk.expr import Secret, wsum_secrets
from zksk.extended import ExtendedProofStmt
from zksk.composition import AndProofStmt
from zksk.primitives.dlrep import DLRep
from zksk.utils import make_generators


@attr.s
class BBSPlusSignature:
    """
    BBS+ signature.
    """

    A = attr.ib()
    e = attr.ib()
    s = attr.ib()

    def verify_signature(self, pk, messages):
        """
        Verify the validity of the signature w.r.t the given public key and set of messages.
        """
        generators = pk.generators[: len(messages) + 2]
        product = generators[0] + generators[0].group.wsum(
            ([self.s] + messages), generators[1:]
        )
        return self.A.pair(pk.w + self.e * pk.h0) == product.pair(pk.h0)


@attr.s
class UserCommitmentMessage:
    """
    Embed the product to be pre-signed by the issuer.

    If blinded by a user's Pedersen commitment, a NI proof is also specified.
    """

    com_message = attr.ib()
    com_nizk_proof = attr.ib(default=None)

    def verify_blinding(self, pk):
        """
        Verify the NIZK proof for Pedersen commitment.
        """
        if self.com_nizk_proof is None:
            raise ValueError("No proof to verify")

        # TODO: Extract into a separate ExtendedProofStmt.
        lhs = self.com_message
        generators = pk.generators[1 : len(self.com_nizk_proof.responses) + 1]
        secret_vars = [Secret() for _ in self.com_nizk_proof.responses]
        proof = DLRep(lhs, wsum_secrets(secret_vars, generators))

        return proof.verify(self.com_nizk_proof)


class BBSPlusSignatureCreator:
    """
    Pre-signed product along with a NIZK proof of correct construction.

    Args:
        pk (PublicKey): Public key.
    """

    def __init__(self, pk):
        self.pk = pk
        self.s1 = None

    def commit(self, messages, zkp=True):
        """
        Construct the product of messages and optionaly a Pedersen commitment and its proof.

        Args:
            messages: Messages (attributes) to commit to
            zkp (bool): Whether to construct a Pedersen commitment and proof the knowledge of the
                messages for this commitment.

        Returns:
            :py:class:`UserCommitmentMessage`: user's packed commitment.
        """
        lhs = self.pk.generators[0].group.wsum(
            messages, self.pk.generators[2 : len(messages) + 2]
        )
        com_nizk_proof = None
        if zkp:
            self.s1 = self.pk.generators[0].group.order().random()
            lhs = self.s1 * self.pk.generators[1] + lhs

            # TODO: Extract into a separate ExtendedProofStmt.
            secret_vars = [Secret() for _ in range(len(messages) + 1)]
            secrets = [self.s1] + messages
            rhs = wsum_secrets(secret_vars, self.pk.generators[1 : len(messages) + 2])
            com_stmt = DLRep(lhs, rhs)
            com_nizk_proof = com_stmt.prove(
                {s: v for s, v in zip(secret_vars, secrets)}
            )

        return UserCommitmentMessage(com_message=lhs, com_nizk_proof=com_nizk_proof)

    def obtain_signature(self, presignature):
        """
        Make a complete signature from the received pre-signature.

        Args:
            presignature (:py:class:`BBSPlusSignature`): presignature

        Returns:
            :py:class:`BBSPlusSignature`: Signature.
        """

        # s1 is the part of the signature blinding factor which is on the user side.
        if self.s1 is None:
            new_s = presignature.s
        else:
            new_s = presignature.s + self.s1
        return BBSPlusSignature(A=presignature.A, e=presignature.e, s=new_s)


@attr.s
class BBSPlusKeypair:
    """
    A public-private key pair, along with a list of canonical bases to use in proofs.
    """

    generators = attr.ib()
    h0 = attr.ib()
    sk = attr.ib()
    pk = attr.ib()

    @staticmethod
    def generate(bilinear_pair, num_generators):
        """
        Generate a keypair.

        Args:
            bilinear_pair (:py:class:`pairings.BilinearGroupPair`): Bilinear group pair.
            num_generators: Upper bound on the number of generators needed to compute the proof.
                Should be at least `2 + the number of messages`.

        Returns:
            :py:class:`BBSPlusKeypair`: Keypair.
        """
        # TODO: Check if this +2 is not redundant.
        generators = make_generators(num_generators + 2, group=bilinear_pair.G1)
        h0 = bilinear_pair.G2.generator()
        sk = BBSPlusSecretKey(
            gamma=bilinear_pair.G1.order().random(), generators=generators, h0=h0,
        )
        pk = BBSPlusPublicKey(w=sk.gamma * h0, generators=generators, h0=h0)
        return BBSPlusKeypair(generators=generators, h0=h0, sk=sk, pk=pk)
        # self.sk.pk = self.pk


@attr.s
class BBSPlusPublicKey:
    """
    BBS+ public key.

    Automatically pre-computes the generator pairings :math:`e(g_i, h_0)`.
    """

    w = attr.ib()
    h0 = attr.ib()
    generators = attr.ib()

    def __attrs_post_init__(self):
        """Pre-compute the group pairings."""
        self.gen_pairs = [g.pair(self.h0) for g in self.generators]


@attr.s
class BBSPlusSecretKey:
    """
    BBS+ private key.
    """

    h0 = attr.ib()
    gamma = attr.ib()
    generators = attr.ib()

    def sign(self, lhs):
        r"""
        Sign a committed message (typically a product, blinded or not),

        A signature is :math:`(A, e, s_2)` such that
        :math:`A = (g_0 + s_2 g_1 + C_m) \cdot \frac{1}{e+\gamma}`.

        If the product was blinded by the user's :math:`s_1` secret value, user has to update the
        signature.
        """
        pedersen_product = lhs
        e = self.h0.group.order().random()
        s2 = self.h0.group.order().random()
        prod = self.generators[0] + s2 * self.generators[1] + pedersen_product
        A = (self.gamma + e).mod_inverse(self.h0.group.order()) * prod
        return BBSPlusSignature(A=A, e=e, s=s2)


class BBSPlusSignatureStmt(ExtendedProofStmt):
    """
    Proof of knowledge of a BBS+ signature over a set of (hidden) messages.

    The proof can be made `binding`: bind the secrets to another proof. If the proof is not binding,
    it is not possible to assert that the same secrets were used in any other proof.

    Args:
        secret_vars: Secret variables.
            If binding, the two first elements of secret_vars as the Secret variables for the ``e``
            and ``s`` attributes of the signature.
        pk (:py:class:`BBSPlusPublicKey`): Public key.
        signature (:py:class:`BBSPlusSignature`): Signature. Required if used for proving.
        binding (bool): Whether the signature is binding.
        simulated (bool): If this proof is a part of an or-proof: whether it should be simulated.
    """

    def __init__(self, secret_vars, pk, signature=None, binding=True, simulated=False):
        self.pk = pk
        self.signature = signature
        if not binding:
            # We add two Secret slots for e and s if necessary
            secret_vars = [Secret(), Secret()] + secret_vars

        # We need L+1 generators for L messages. secret_vars are messages plus 'e' and 's'
        self.bases = pk.generators[: len(secret_vars)]
        self.order = self.bases[0].group.order()

        # The prover will compute the following secrets:
        self.r1, self.r2, self.delta1, self.delta2 = (
            Secret(),
            Secret(),
            Secret(),
            Secret(),
        )

        # Below is boilerplate
        self.secret_vars = secret_vars
        if signature is not None:
            # Digest the signature parameters
            self.secret_vars[0].value = signature.e
            self.secret_vars[1].value = signature.s

        self.set_simulated(simulated)

    def precommit(self):
        """
        Generate the lacking information to construct a complete proof.

        The precommitment comprises the ``A1`` and ``A2`` commitments that depend on the secret
        signature and the Prover's randomness.
        """
        if self.signature is None:
            raise ValueException("No signature given!")

        # Compute auxiliary commitments A1,A2 as mentioned in the paper. Needs two random values r1,r2 and associated delta1,delta2
        # Set true value to computed secrets
        r1, r2 = self.order.random(), self.order.random()
        self.r1.value, self.r2.value = r1, r2
        self.delta1.value = r1 * self.signature.e % self.order
        self.delta2.value = r2 * self.signature.e % self.order

        precommitment = {}
        precommitment["A1"] = r1 * self.bases[1] + r2 * self.bases[2]
        precommitment["A2"] = r1 * self.bases[2] + self.signature.A

        return precommitment

    def construct_stmt(self, precommitment):
        r"""
        Proof of knowledge of a signature.

        This is an implementation of a proof :math:`\Pi_5` detailed on page 7 of the `Constant-Size
        Dynamick-TAA` paper.
        """

        self.A1, self.A2 = precommitment["A1"], precommitment["A2"]
        g0, g1, g2 = self.bases[0], self.bases[1], self.bases[2]

        dl1 = DLRep(self.A1, self.r1 * g1 + self.r2 * g2)
        dl2 = DLRep(
            g0.group.infinite(),
            self.delta1 * g1 + self.delta2 * g2 + self.secret_vars[0] * (-1 * self.A1),
        )

        self.pair_lhs = self.A2.pair(self.pk.w) + (-1 * self.pk.gen_pairs[0])
        bases = [
            -1 * (self.A2.pair(self.pk.h0)),
            self.bases[2].pair(self.pk.w),
            self.pk.gen_pairs[2],
        ]
        bases.extend(self.pk.gen_pairs[1 : len(self.bases)])

        # Build secret names [e, r1, delta1, s, m_i]
        new_secret_vars = (
            self.secret_vars[:1] + [self.r1, self.delta1] + self.secret_vars[1:]
        )
        pairings_stmt = DLRep(self.pair_lhs, wsum_secrets(new_secret_vars, bases))

        constructed_stmt = AndProofStmt(dl1, dl2, pairings_stmt)
        constructed_stmt.lhs = [p.lhs for p in constructed_stmt.subproofs]
        return constructed_stmt

    def simulate_precommit(self):
        """
        Draw :math:`A_1`, :math:`A_2` at random.
        """
        group = self.bases[0].group

        precommitment = {}
        precommitment["A1"] = group.order().random() * group.generator()
        precommitment["A2"] = group.order().random() * group.generator()
        return precommitment



"""
Common clases, including subclassable basic provers and verifiers.
"""

import abc
import random
import string
from petlib.ec import EcGroup, EcPt
from petlib.bn import Bn
from petlib.pack import *
import binascii
import msgpack
from hashlib import sha256
from collections import defaultdict
import attr

from zksk.utils import get_random_num
from zksk.consts import CHALLENGE_LENGTH
from zksk.exceptions import ValidationError


@attr.s
class NIZK:
    """
    Non-interactive zero-knowledge proof.
    """

    challenge = attr.ib()
    responses = attr.ib()
    precommitment = attr.ib(default=None)
    stmt_hash = attr.ib(default=None)


    def serialize(self):
        """
        Serialize a non-interactive zero-knowledge proof.
        """
        as_list = [
            encode(self.challenge),
            encode(self.responses),
            encode(self.precommitment),
            encode(self.stmt_hash)
        ]
        return msgpack.packb(as_list, use_bin_type=True)


    @classmethod
    def deserialize(cls, nizk_raw):
        """
        Deserialize a non-interactive zero-knowledge proof.
        """
        as_list = [decode(x) for x in msgpack.unpackb(nizk_raw)]
        return NIZK(*as_list)


@attr.s
class SimulationTranscript:
    """
    Simulated proof transcript.
    """

    commitment = attr.ib()
    challenge = attr.ib()
    responses = attr.ib()
    precommitment = attr.ib(default=None)
    stmt_hash = attr.ib(default=None)


def build_fiat_shamir_challenge(stmt_prehash, *args, message=""):
    """Generate a Fiat-Shamir challenge.

    >>> prehash = sha256(b"statement id")
    >>> commitment = 42 * EcGroup().generator()
    >>> isinstance(build_fiat_shamir_challenge(prehash, commitment), Bn)
    True

    Args:
        prehash: Hash object seeded with the proof statement ID.
        args: Items to hash (e.g., commitments)
        message: Message to make it a signature PK.
    """
    # Start building the complete hash for the challenge
    for elem in args:
        if not isinstance(elem, bytes) and not isinstance(elem, str):
            encoded = encode(elem)
        else:
            encoded = elem
        stmt_prehash.update(encoded)

    stmt_prehash.update(message.encode())
    return Bn.from_hex(stmt_prehash.hexdigest())


class Prover(metaclass=abc.ABCMeta):
    """
    Abstract interface representing Prover used in sigma protocols.

    Args:
        stmt: The Proof instance from which we draw the Prover.
        secret_values: The values of the secrets as a dict.
    """

    def __init__(self, stmt, secret_values):
        self.stmt = stmt
        self.secret_values = secret_values

    @abc.abstractmethod
    def compute_response(self, challenge):
        """
        Computes the responses associated to each Secret object in the statement.

        Returns a list of responses.
        """
        pass

    def precommit(self):
        """
        Generate a precommitment.
        """
        return None

    def commit(self, randomizers_dict=None):
        """
        Constuct the proof commitment.

        Args:
            randomizers_dict: Optional dictionary of random values. Each random values is assigned
                to a secret.
        """
        return (
            self.stmt.prehash_statement().digest(),
            self.internal_commit(randomizers_dict),
        )

    def get_nizk_proof(self, message=""):
        """
        Construct a non-interactive proof transcript using Fiat-Shamir heuristic.

        The transcript contains only the challenge and the responses, as the commitment can be
        deterministically recomputed.

        The challenge is a hash of the commitment, the stmt statement and all the bases in the
        statement (including the left-hand-side).

        Args:
            message (str): Optional message to make a signature stmt of knowledge.
        """
        # Precommit to gather encapsulated precommitments. They are already included in their
        # respective statement.
        precommitment = self.precommit()
        commitment = self.internal_commit()

        # Generate the challenge.
        prehash = self.stmt.prehash_statement()
        stmt_hash = prehash.digest()
        challenge = build_fiat_shamir_challenge(
            prehash, precommitment, commitment, message=message
        )

        responses = self.compute_response(challenge)
        return NIZK(
            challenge=challenge,
            responses=responses,
            precommitment=precommitment,
            stmt_hash=stmt_hash,
        )


class Verifier(metaclass=abc.ABCMeta):
    """
    An abstract interface representing Prover used in sigma protocols
    """

    def __init__(self, stmt):
        self.stmt = stmt

    @abc.abstractmethod
    def check_responses_consistency(self, response, response_dict=None):
        """
        Verify that for two identical secrets, the responses are also the same.
        """
        return False

    def process_precommitment(self, precommitment):
        """
        Receive a precommitment and process it.
        """
        pass

    def send_challenge(self, commitment):
        """
        Store the received commitment and generate a challenge.

        The challenge is chosen at random between 0 and ``CHALLENGE_LENGTH`` (excluded).

        Args:
            commitment: A tuple containing a hash of the stmt statement, to be
                compared against the local statement, and the commmitment as a
                (potentially multi-level list of) base(s) of the group.
        """
        statement, self.commitment = commitment
        self.stmt.check_statement(statement)
        self.challenge = get_random_num(bits=CHALLENGE_LENGTH)
        return self.challenge

    def pre_verification_validation(self, response, *args, **kwargs):
        self.stmt.full_validate(*args, **kwargs)

        if not self.check_responses_consistency(response, {}):
            raise ValidationError("Responses for the same secret name do not match.")

    def verify(self, response, *args, **kwargs):
        """
        Verify the responses of an interactive sigma protocol.

        To do so, generates a pseudo-commitment based on the stored challenge and the received
        responses, and compares it against the stored commitment.

        Args:
            response: The response given by the prover

        Returns:
            bool: True if verification succeeded, False otherwise.
        """
        # TODO: I really don't think this chain should be raising exceptions
        self.pre_verification_validation(response, *args, **kwargs)

        # Retrieve the commitment using the verification identity
        return self.commitment == self.stmt.recompute_commitment(
            self.challenge, response
        )

    def verify_nizk(self, nizk, message="", *args, **kwargs):
        """
        Verify a non-interactive proof.

        Unpacks the attributes and checks their consistency by computing a pseudo-commitment and
        drawing from a pseudo-challenge. Compares the pseudo-challenge with the nizk challenge.

        Args:
            nizk (:py:class:`NIZK`): Non-interactive proof
            message: A message if a signature proof.

        Return:
            bool: True of verification succeeded, False otherwise.
        """
        # Build the complete stmt if necessary.
        # TODO: If empty precommit() function, this is always true.
        if nizk.precommitment is not None:
            self.process_precommitment(nizk.precommitment)

        # Check the proofs statements match, gather the local statement.
        prehash = self.stmt.check_statement(nizk.stmt_hash)
        self.pre_verification_validation(nizk.responses, *args, **kwargs)

        # Retrieve the commitment using the verification identity.
        commitment_prime = self.stmt.recompute_commitment(
            nizk.challenge, nizk.responses
        )
        challenge_prime = build_fiat_shamir_challenge(
            prehash, nizk.precommitment, commitment_prime, message=message
        )
        return nizk.challenge == challenge_prime



# TODO: Fix the disastrous variable naming.
# TODO: Split into more granular unit tests.

import pytest

from bplib.bp import BpGroup
from petlib import pack

from zksk.pairings import BilinearGroupPair, G1Point, AdditivePoint


@pytest.fixture
def bp_group():
    return BpGroup()


@pytest.fixture
def group_pair():
    return BilinearGroupPair()


def test_gt_order(bp_group, group_pair):
    """
    Given a group GT defining a pairing e(G1,G2)->GT, GTGRoup behaves similarly.
    """
    GT = group_pair.GT
    assert GT == group_pair.groups()[2]
    assert bp_group.order() == GT.order()


def test_g1_point(bp_group, group_pair):
    """
    G1Point class which overrides G1 and G2 points in such a way
    that pt.group actually returns the G1/G2 groups and not the GT. Also, these
    points allow internal pair.
    """
    g = bp_group.pair(bp_group.gen1(), bp_group.gen2())
    gmg = group_pair.GT.generator()
    assert g == gmg.pt
    assert bp_group.pair(bp_group.gen1(), bp_group.gen2()) == gmg.pt
    assert g == group_pair.G1.generator().pair(group_pair.G2.generator()).pt


def test_additive_point(bp_group, group_pair):
    """
    AdditivePoint class defines GT points additively
    """
    g = bp_group.pair(bp_group.gen1(), bp_group.gen2())
    gmg = group_pair.GT.generator()
    assert g == gmg.pt
    assert gmg == AdditivePoint(g, group_pair)
    assert AdditivePoint(g ** (g.group.order()), group_pair) == group_pair.GT.infinite()

    r = bp_group.order().random()
    g1, g1mg = g ** r, r * gmg
    assert g1 == g1mg.pt
    assert g1 * g1 * g1 == (g1mg + g1mg + g1mg).pt
    assert g1.export() == g1mg.export()
    assert g1.group == g1mg.bp.bpgp


def test_g1_g2_groups(bp_group, group_pair):
    G1, G2, GT = group_pair.groups()
    g1, g2 = G1.generator(), G2.generator()
    assert G1.infinite().pt == bp_group.gen1().inf(bp_group)
    assert G1.infinite() == G1Point(bp_group.gen1().inf(bp_group), group_pair)
    assert g1 * 0 == G1.infinite()
    assert g1 * 0 == g1 * bp_group.order()
    assert G2.infinite().pt == bp_group.gen2().inf(bp_group)
    assert g2 * 0 == G2.infinite()
    assert g2 * 0 == g2 * bp_group.order()


def test_pack_unpack_g1(group_pair):
    """
    Testing packing and unpacking G1 element
    """
    group_pair = BilinearGroupPair()
    g1 = group_pair.G1
    order = g1.order()
    pt1 = order.random() * g1.generator()

    data = pack.encode(pt1)
    pt2 = pack.decode(data)

    assert pt1 == pt2


def test_pack_unpack_g2(group_pair):
    """
    Testing packing and unpacking G2 element
    """
    group_pair = BilinearGroupPair()
    g2 = group_pair.G2
    order = g2.order()
    pt1 = order.random() * g2.generator()

    data = pack.encode(pt1)
    pt2 = pack.decode(data)

    assert pt1 == pt2


def test_pack_unpack_gt(group_pair):
    """
    Testing packing and unpacking GT element
    """
    group_pair = BilinearGroupPair()
    gt = group_pair.GT
    order = gt.order()
    pt1 = order.random() * gt.generator()

    data = pack.encode(pt1)
    pt2 = pack.decode(data)

    assert pt1 == pt2



import math
import secrets
import hashlib
import warnings

from petlib.bn import Bn

from zksk.consts import DEFAULT_GROUP
from zksk.exceptions import InvalidExpression


def get_random_point(group=None, random_bits=256, seed=None):
    """
    Generate some random group generators.

    Args:
        num: Number of generators to generate.
        group: Group
        random_bits: Number of bits of a random string to create a point.

    >>> from petlib.ec import EcPt
    >>> a = get_random_point()
    >>> b = get_random_point()
    >>> isinstance(a, EcPt)
    True
    >>> isinstance(b, EcPt)
    True
    >>> a != b
    True
    >>> get_random_point(seed=1)
    EcPt(037697679766c26bb7b76c65d2639fb983dea7c859c63b3047168dbc1b)
    """

    if group is None:
        group = DEFAULT_GROUP

    num_bytes = math.ceil(random_bits / 8)
    if seed is None:
        randomness = secrets.token_bytes(num_bytes)
    else:
        randomness = hashlib.sha512(b"%i" % seed).digest()[:num_bytes]

    return group.hash_to_point(randomness)


def make_generators(num, group=None, random_bits=256, seed=42):
    """
    Create some random group generators.

    .. WARNING ::

        There is a negligible chance that some generators will be the same.

    Args:
        num: Number of generators to generate.
        group: Group
        random_bits: Number of bits of a random number used to create a generator.

    >>> from petlib.ec import EcPt
    >>> generators = make_generators(3)
    >>> len(generators) == 3
    True
    >>> isinstance(generators[0], EcPt)
    True
    """
    if group is None:
        group = DEFAULT_GROUP
    generators = [
        get_random_point(
            group, random_bits, seed=seed + i if seed is not None else None
        )
        for i in range(num)
    ]
    return generators


def get_random_num(bits):
    """
    Draw a random number of given bitlength.

    >>> x = get_random_num(6)
    >>> x < 2**6
    True
    """
    order = Bn(2).pow(bits)
    return order.random()


def sum_bn_array(arr, modulus):
    """
    Sum an array of big numbers under a modulus.

    >>> a = [Bn(5), Bn(7)]
    >>> m = 10
    >>> sum_bn_array(a, m)
    2
    """
    if not isinstance(modulus, Bn):
        modulus = Bn(modulus)
    res = Bn(0)
    for elem in arr:
        if not isinstance(elem, Bn):
            elem = Bn(elem)
        res = res.mod_add(elem, modulus)
    return res


def ensure_bn(x):
    """
    Ensure that value is big number.

    >>> isinstance(ensure_bn(42), Bn)
    True
    >>> isinstance(ensure_bn(Bn(42)), Bn)
    True
    """
    if isinstance(x, Bn):
        return x
    else:
        return Bn(x)



import os
import sys
import importlib

import pytest

PACKAGE_NAME = "zksk"
EXAMPLES_MOD = "examples"
BASE_PATH = dir_path = os.path.dirname(os.path.realpath(__file__))
EXAMPLES_DIR = os.path.join(BASE_PATH, "..", EXAMPLES_MOD)
EXAMPLE_MODULE_NAMES = [f[:-3] for f in os.listdir(EXAMPLES_DIR) if f.endswith(".py")]


@pytest.mark.parametrize("mod", EXAMPLE_MODULE_NAMES)
def test_examples(mod):
    importlib.import_module("%s.%s" % (EXAMPLES_MOD, mod), PACKAGE_NAME)



"""
Composable proof statements, provers, and verifiers.
"""

import abc
import copy
import random
from hashlib import sha256
from collections import defaultdict

from petlib.bn import Bn
from petlib.pack import encode

from zksk.consts import CHALLENGE_LENGTH
from zksk.base import Prover, Verifier, SimulationTranscript
from zksk.expr import Secret, update_secret_values
from zksk.utils import get_random_num, sum_bn_array
from zksk.utils.misc import get_default_attr
from zksk.exceptions import StatementSpecError, StatementMismatch
from zksk.exceptions import InvalidSecretsError, GroupMismatchError
from zksk.exceptions import InconsistentChallengeError


def _find_residual_challenge(subchallenges, challenge, modulus):
    """
    Determine the complement to a global challenge in a list

    For example, to find :math:`c_1` such that :math:`c = c_1 + c_2 + c_3 \mod k`, we compute
    :math:`c_2 + c_3 - c` and take the opposite.

    Args:
        subchallenges: The array of subchallenges :math:`c_2`, c_3, ...`
        challenge: The global challenge to reach
        modulus: the modulus :math:`k`
    """
    modulus = Bn(2).pow(modulus)
    temp_arr = subchallenges.copy()
    temp_arr.append(-challenge)
    return -sum_bn_array(temp_arr, modulus)


def _assign_secret_ids(secret_vars):
    """
    Assign consecutive identifiers to secrets.

    Needed for proof statement idenfifiers.

    >>> x, y = Secret(name='x'), Secret(name='y')
    >>> _assign_secret_ids([x, y, x])
    {'x': 0, 'y': 1}

    Args:
        secret_vars: :py:class:`expr.Secret` objects.
    """
    secret_id_map = {}
    counter = 0
    for secret in secret_vars:
        if secret.name not in secret_id_map:
            secret_id_map[secret.name] = counter
            counter += 1
    return secret_id_map


class ComposableProofStmt(metaclass=abc.ABCMeta):
    """
    A composable sigma-protocol proof statement.

    In the composed proof tree, these objects are the atoms/leafs.
    """

    def get_proof_id(self, secret_id_map=None):
        """
        Identifier for the proof statement.

        This identifier is used to check the proof statements on the prover and
        verifier sides are consistent, and to generate a challenge in non-interactive proofs.

        Args:
            secret_id_map: A map from secret names to consecutive identifiers.

        Returns:
            list: Objects that can be used for hashing.
        """
        secret_vars = self.get_secret_vars()
        bases = self.get_bases()
        if secret_id_map is None:
            secret_id_map = _assign_secret_ids(secret_vars)
        ordered_secret_ids = [secret_id_map[s.name] for s in secret_vars]
        return [self.__class__.__name__, bases, ordered_secret_ids]

    def get_secret_vars(self):
        """
        Collect all secrets in this subtree.

        By default tries to get the ``secret_vars`` attribute. Override if needed.
        """
        if not hasattr(self, "secret_vars"):
            raise StatementSpecError(
                "Need to override get_secret_vars or specify secret_vars attribute."
            )
        return self.secret_vars

    def get_bases(self):
        """
        Collect all base points in this subtree.

        By default tries to get the ``bases`` attribute. Override if needed.
        """
        if not hasattr(self, "bases"):
            raise StatementSpecError(
                "Need to override get_bases or specify bases attribute."
            )
        return self.bases

    def __and__(self, other):
        """
        Make a conjuction of proof statements using :py:class:`AndProofStmt`.

        If called multiple times, subproofs are flattened so that only one :py:class:`AndProofStmt`
        remains at the root.
        """
        if isinstance(other, AndProofStmt):
            if isinstance(self, AndProofStmt):
                return AndProofStmt(*self.subproofs, *other.subproofs)
            else:
                return AndProofStmt(self, *other.subproofs)

        elif isinstance(self, AndProofStmt):
            return AndProofStmt(*self.subproofs, other)

        return AndProofStmt(self, other)

    def __or__(self, other):
        """
        Make a disjunction of proof statements using :py:class:`OrProofStmt`.

        If called multiple times, subproofs are flattened so that only one :py:class:`OrProofStmt`
        remains at the root.
        """
        if isinstance(other, OrProofStmt):
            if isinstance(self, OrProofStmt):
                return OrProofStmt(*self.subproofs, *other.subproofs)
            else:
                return OrProofStmt(self, *other.subproofs)

        elif isinstance(self, OrProofStmt):
            return OrProofStmt(*self.subproofs, other)

        return OrProofStmt(self, other)

    def get_prover_cls(self):
        if hasattr(self, "prover_cls"):
            return self.prover_cls
        else:
            raise StatementSpecError("No prover class specified.")

    def get_verifier_cls(self):
        if hasattr(self, "verifier_cls"):
            return self.verifier_cls
        else:
            raise StatementSpecError("No verifier class specified.")

    def get_prover(self, secrets_dict=None):
        """
        Get the :py:class:`base.Prover` for the current proof.
        """
        return self.get_prover_cls()(self)

    def get_verifier(self):
        """
        Return the :py:class:`base.Verifier` for the current proof.
        """
        return self.get_verifier_cls()(self)

    def recompute_commitment(self, challenge, response):
        """
        Compute a pseudo-commitment.

        A pseudo-commitment is the commitment a verifier should have received if the proof was
        correct. It should be compared to the actual commitment.

        Re-occuring secrets yield identical responses.

        Args:
            challenge: the challenge used in the proof
            response: a list of responses, ordered as the list of secret names, i.e., with as many
                elements as there are secrets in the proof claim.
        """
        pass

    def prove(self, secret_dict=None, message=""):
        """
        Generate the transcript of a non-interactive proof.
        """
        if secret_dict is None:
            secret_dict = {}
        prover = self.get_prover(secret_dict)
        return prover.get_nizk_proof(message)

    def verify(self, nizk, message=""):
        """
        Verify a non-interactive proof.
        """
        verifier = self.get_verifier()
        return verifier.verify_nizk(nizk, message)

    def check_statement(self, statement_hash):
        """
        Verify the current proof corresponds to the hash passed as a parameter.

        Returns a pre-hash of the current proof, e.g., to be used to verify NIZK proofs.
        """
        h = self.prehash_statement()
        if statement_hash != h.digest():
            raise StatementMismatch("Proof statements mismatch, impossible to verify")
        return h

    def validate(self, *args, **kwargs):
        """
        Validation criteria to be checked. Override if needed.

        For example, a :py:class:`primitives.DLNotEqual` statement should
        not validate if its proof components are in fact equal.

        Should raise an `exceptions.ValidationError` if does not validate.
        """
        pass

    def full_validate(self, *args, **kwargs):
        """
        For or/and-proofs, perform recursive validation of subproofs.
        """
        # TODO: calling return here is deceptive if we ask to throw an Exception
        return self.validate(*args, **kwargs)

    def validate_secrets_reoccurence(self, forbidden_secrets=None):
        """
        Check if a secret appears both inside an outside an or-proof.

        Does nothing if not overriden.
        """
        pass

    def update_randomizers(self, randomizers_dict):
        """
        Construct a mapping of all secrets to randomizers.

        Does so by copying the values of the passed ``randomizers_dict``, and drawing the other
        values at random until all the secrets have a randomizer.

        These are used as a part of proofs and also as responses in simulations.

        Args:
            randomizers_dict: A dictionary to enforce
        """
        # If we are not provided a randomizer dict from above, we compute it.
        if randomizers_dict is None:
            randomizers_dict = self.get_randomizers()

        # TODO: This can be done easier.

        # Fill the dictionary.
        elif any([x not in randomizers_dict for x in self.get_secret_vars()]):
            tmp = self.get_randomizers()
            tmp.update(randomizers_dict)
            randomizers_dict = tmp

        return randomizers_dict

    def prehash_statement(self):
        """
        Return a hash of the proof's ID.
        """
        return sha256(encode(str(self.get_proof_id())))

    @property
    def simulated(self):
        """
        Tell if this proof is designated as to be simulated in an or-proof.

        By default is False.
        """
        return get_default_attr(self, "_simulated", False)

    def set_simulated(self, value=True):
        """
        Designate this proof statement as simulated in an or-proof.

        Args:
            value (bool): Whether to simulate this proof.
        """
        self._simulated = value

    def prepare_simulate_proof(self):
        """
        Additional steps to prepare before simulating the proof. Override if needed.
        """
        pass

    def simulate(self, challenge=None):
        """
        Generate the transcript of a simulated non-interactive proof.
        """
        self.set_simulated()
        self.prepare_simulate_proof()
        transcript = self.simulate_proof(challenge=challenge)
        transcript.stmt_hash = self.prehash_statement().digest()
        return transcript

    def verify_simulation_consistency(self, transcript):
        """Check if the fields of a transcript satisfy the verification equation.

        Useful for debugging purposes.

        .. WARNING::

            This is NOT an alternative to the full proof verification, as this function
            accepts simulated proofs.

        """
        verifier = self.get_verifier()
        verifier.process_precommitment(transcript.precommitment)
        self.check_statement(transcript.stmt_hash)
        verifier.commitment, verifier.challenge = (
            transcript.commitment,
            transcript.challenge,
        )
        return verifier.verify(transcript.responses)

    def __repr__(self):
        # TODO: Not a great repr (cannot copy-paste and thus recreate the object).
        return str(self.get_proof_id())


class _CommonComposedStmtMixin:
    def get_secret_vars(self):
        secret_vars = []
        for sub in self.subproofs:
            secret_vars.extend(sub.get_secret_vars())
        return secret_vars

    def get_bases(self):
        bases = []
        for sub in self.subproofs:
            bases.extend(sub.get_bases())
        return bases

    def validate_group_orders(self):
        """
        Check that if two secrets are the same, their bases induce groups of the same order.

        The primary goal is to ensure same responses for same secrets will not yield false negatives
        of :py:meth:`base.Verifier.check_responses_consistency` due to different group-order modular
        reductions.

        TODO: Consider deactivating in the future as this forbids using different groups in one proof.
        TODO: Update docs, variable names.

        Args:
            secrets: :py:class:`expr.Secret` objects.
            bases: Elliptic curve base points.
        """
        bases = self.get_bases()
        secrets = self.get_secret_vars()

        # We map the unique secrets to the indices where they appear
        mydict = defaultdict(list)
        for index, word in enumerate(secrets):
            mydict[word].append(index)

        # Now we use this dictionary to check all the bases related to a particular secret live in
        # the same group
        for (word, gen_idx) in mydict.items():
            # Word is the key, gen_idx is the value = a list of indices
            ref_order = bases[gen_idx[0]].group.order()

            for index in gen_idx:
                if bases[index].group.order() != ref_order:
                    raise GroupMismatchError(
                        "A shared secret has bases which yield different group orders: %s"
                        % word
                    )

    def get_proof_id(self, secret_id_map=None):
        secret_vars = self.get_secret_vars()
        bases = self.get_bases()
        if secret_id_map is None:
            secret_id_map = _assign_secret_ids(secret_vars)

        proof_ids = [sub.get_proof_id(secret_id_map) for sub in self.subproofs]
        return (self.__class__.__name__, proof_ids)

    def full_validate(self, *args, **kwargs):
        for sub in self.subproofs:
            sub.full_validate(*args, **kwargs)


class OrProofStmt(_CommonComposedStmtMixin, ComposableProofStmt):
    """
    An disjunction of several subproofs.


    Args:
        suproofs: Proof statements.

    Raise:
        ValueError: If less than two subproofs given.
    """

    def __init__(self, *subproofs):
        if len(subproofs) < 2:
            raise ValueError("Need at least two subproofs")

        # We make a shallow copy of each subproof so they don't mess up each other.  This step is
        # important, as we can have different outputs for the same proof (independent simulations or
        # simulations/execution)
        self.subproofs = [copy.copy(p) for p in list(subproofs)]

    def recompute_commitment(self, challenge, responses):
        # We retrieve the challenges, hidden in the responses tuple
        self.or_challenges = responses[0]
        responses = responses[1]

        # We check for challenge consistency i.e the constraint was respected
        if _find_residual_challenge(
            self.or_challenges, challenge, CHALLENGE_LENGTH
        ) != Bn(0):
            raise InconsistentChallengeError("Inconsistent challenges.")

        # Compute the list of commitments, one for each proof with its challenge and responses
        # (in-order)
        com = []
        for index, subproof in enumerate(self.subproofs):
            com.append(
                subproof.recompute_commitment(
                    self.or_challenges[index], responses[index]
                )
            )
        return com

    def get_prover(self, secrets_dict=None):
        if secrets_dict is None:
            secrets_dict = {}

        # The prover is built on one legit prover constructed from a subproof picked at random among
        # candidates.
        # First we update the dictionary we have with the additional secrets, and process it
        # TODO: Check this secret_values handling totally different
        update_secret_values(secrets_dict)

        if self.simulated:
            return None

        # TODO: Add a unit test where simulation must be True/False for all subproofs

        # Prepare the draw. Disqualify proofs with simulation parameter set to true
        candidates = {}
        for index, subproof in enumerate(self.subproofs):
            if not self.subproofs[index].simulated:
                candidates[index] = subproof

        if len(candidates) == 0:
            print("Cannot run an or-proof if all elements are simulated")
            return None

        # Now choose a proof among the possible ones and try to get a prover from it.
        # If for some reason it does not work (e.g some secrets are missing), remove it
        # from the list of possible proofs and try again
        random_gen = random.SystemRandom()
        possible = list(candidates.keys())
        self.chosen_idx = random_gen.choice(possible)

        # Feed the selected proof the secrets it needs if we have them, and try to get_prover
        valid_prover = self.subproofs[self.chosen_idx].get_prover(secrets_dict)
        while valid_prover is None:
            possible.remove(self.chosen_idx)
            # If there is no proof left, abort and say we cannot get a prover
            if len(possible) == 0:
                self.chosen_idx = None
                return None
            self.chosen_idx = random_gen.choice(possible)
            valid_prover = self.subproofs[self.chosen_idx].get_prover(secrets_dict)
        return OrProver(self, valid_prover)

    def get_verifier(self):
        return OrVerifier(self, [sub.get_verifier() for sub in self.subproofs])

    def validate_composition(self):
        """
        Validate that composition is done correctly.
        """
        self.validate_group_orders()

    def validate_secrets_reoccurence(self, forbidden_secrets=None):
        """
        Check for re-occurence of secrets both inside and outside an or-proof.

        Method is called from :py:meth:`AndProofStmt.validate_secrets_reoccurence`.

        Args:
            forbidden_secrets: A list of all the secrets in the mother proof.

        Raises:
            :py:class:`exceptions.InvalidSecretsError`: If any secrets re-occur in an
                unsupported way.
        """
        secret_vars = self.get_secret_vars()
        if forbidden_secrets is None:
            return

        for secret in set(secret_vars):
            if forbidden_secrets.count(secret) > secret_vars.count(secret):
                raise InvalidSecretsError(
                    "Invalid secrets found. Try to flatten the proof to avoid "
                    "using secrets used inside an or-proof in other parts of "
                    "the proof too (e.g., in other and or or-clauses)"
                )

    def prepare_simulate_proof(self):
        for sub in self.subproofs:
            sub.prepare_simulate_proof()

    def simulate_proof(self, challenge=None, *args, **kwargs):
        # Simulate the n-1 first subproofs, computes the complementary challenge and
        # simulates the last proof using this challenge.
        if challenge is None:
            challenge = get_random_num(bits=CHALLENGE_LENGTH)
        com = []
        resp = []
        or_chals = []
        precom = []

        # Generate one simulation at a time and update a list of each attribute.
        for index, subproof in enumerate(self.subproofs[:-1]):
            transcript = subproof.simulate_proof()
            com.append(transcript.commitment)
            resp.append(transcript.responses)
            or_chals.append(transcript.challenge)
            precom.append(transcript.precommitment)

        # Generate the last simulation.
        final_chal = _find_residual_challenge(or_chals, challenge, CHALLENGE_LENGTH)
        or_chals.append(final_chal)
        final_transcript = self.subproofs[index + 1].simulate_proof(
            challenge=final_chal
        )
        com.append(final_transcript.commitment)
        resp.append(final_transcript.responses)
        precom.append(final_transcript.precommitment)

        # Pack everything into a SimulationTranscript, pack the or-challenges in the response field.
        return SimulationTranscript(
            commitment=com,
            challenge=challenge,
            responses=(or_chals, resp),
            precommitment=precom,
        )


class OrProver(Prover):
    """
    Prover for the or-proof.

    This prover is built with only one subprover, and needs to have access to the index of the
    corresponding subproof in its mother proof. Runs all the simulations for the other proofs and
    stores them.
    """

    def __init__(self, stmt, subprover):
        self.subprover = subprover
        self.stmt = stmt
        self.true_prover_idx = self.stmt.chosen_idx

        # Create a list storing the SimulationTranscripts
        self.setup_simulations()

    def setup_simulations(self):
        """
        Run all the required simulations and stores them.
        """
        self.simulations = []
        for index, subproof in enumerate(self.stmt.subproofs):
            if index != self.true_prover_idx:
                subproof.prepare_simulate_proof()
                sim = subproof.simulate_proof()
                self.simulations.append(sim)

    def precommit(self):
        # Generate precommitment for the legit subprover, and gather the precommitments from the
        # stored simulations.
        precommitment = []
        for index, _ in enumerate(self.stmt.subproofs):
            if index == self.true_prover_idx:
                precommitment.append(self.subprover.precommit())
            else:
                if index > self.true_prover_idx:
                    index1 = index - 1
                else:
                    index1 = index
                precommitment.append(self.simulations[index1].precommitment)
        if not any(precommitment):
            return None
        return precommitment

    def internal_commit(self, randomizers_dict=None):
        """
        Gather the commitments from the stored simulations.

        Args:
            randomizers_dict: A dictionary of randomizers to use for responses consistency. Not used
                in this proof. Parameter kept so all internal_commit methods have the same prototype.
        """
        # Now that all proofs have been constructed, we can check
        self.stmt.validate_composition()

        commitment = []
        for index, _ in enumerate(self.stmt.subproofs):
            if index == self.true_prover_idx:
                commitment.append(self.subprover.internal_commit())
            else:
                if index > self.true_prover_idx:
                    index1 = index - 1
                else:
                    index1 = index
                commitment.append(self.simulations[index1].commitment)
        return commitment

    def compute_response(self, challenge):
        """
        Compute complementary challenges and responses.

        Computes the complementary challenge with respect to the received global challenge and the
        list of challenges used in the stored simulations.  Computes the responses of the subprover
        using this auxiliary challenge, gathers the responses from the stored simulations.  Returns
        both the complete list of subchallenges (including the auxiliary challenge) and the list of
        responses, both ordered.

        Args:
            challenge: The global challenge to use. All subchallenges must add to this one.
        """
        residual_chal = _find_residual_challenge(
            [el.challenge for el in self.simulations], challenge, CHALLENGE_LENGTH
        )
        response = []
        challenges = []
        for index, subproof in enumerate(self.stmt.subproofs):
            if index == self.true_prover_idx:
                challenges.append(residual_chal)
                response.append(self.subprover.compute_response(residual_chal))
            else:
                # Note that len(simulations) = len(subproofs) - 1.
                if index > self.true_prover_idx:
                    index1 = index - 1
                else:
                    index1 = index
                challenges.append(self.simulations[index1].challenge)
                response.append(self.simulations[index1].responses)

        return (challenges, response)


class OrVerifier(Verifier):
    """
    Verifier for the or-proof.

    The verifier is built on a list of subverifiers, which will unpack the received attributes.
    """

    def __init__(self, stmt, subverifiers):
        self.subs = subverifiers
        self.stmt = stmt

    def process_precommitment(self, precommitment):
        """
        Reads the received list of precommitments (or None if non applicable) and distributes them
        to the subverifiers so they can finalize their proof construction if necessary.

        Args:
            precommitment: A list of all required precommitments, ordered.
        """
        if precommitment is None:
            return
        for index, sub in enumerate(self.subs):
            sub.process_precommitment(precommitment[index])

    def check_responses_consistency(self, responses, responses_dict=None):
        """
        Checks that for a same secret, response are actually the same.

        Since every member is run with its own challenge, it is enough that one
        member is consistent within itself.

        Args:
            responses: a tuple (subchallenges, actual_responses) from which we extract only the
                actual responses for each subverifier.
        """
        if responses_dict is None:
            responses_dict = {}
        for index, sub in enumerate(self.subs):
            if not sub.check_responses_consistency(responses[1][index], {}):
                return False
        return True


class AndProofStmt(_CommonComposedStmtMixin, ComposableProofStmt):
    def __init__(self, *subproofs):
        """
        Constructs the And conjunction of several subproofs.
        Subproofs are copied at instantiation.

        Args:
            suproofs: Proof statements.

        Raise:
            ValueError: If less than two subproofs given.
        """
        if len(subproofs) < 2:
            raise ValueError("Need at least two subproofs")

        # We make a shallow copy of each subproof so they dont mess with each other.  This step is
        # important in case we have proofs which locally draw random values.  It ensures several
        # occurrences of the same proof in the tree indeed have their own randomnesses.
        self.subproofs = [copy.copy(p) for p in list(subproofs)]

    def validate_composition(self, *args, **kwargs):
        """
        Validate that composition is done correctly.
        """
        self.validate_group_orders()
        self.validate_secrets_reoccurence()

    def recompute_commitment(self, challenge, responses):
        com = []
        for index, subproof in enumerate(self.subproofs):
            com.append(subproof.recompute_commitment(challenge, responses[index]))
        return com

    def get_prover(self, secrets_dict=None):
        if secrets_dict is None:
            secrets_dict = {}

        # First we update the dictionary we have with the additional secrets, and process it
        update_secret_values(secrets_dict)

        if self.simulated:
            return None

        subs = [sub_proof.get_prover(secrets_dict) for sub_proof in self.subproofs]

        if None in subs:
            # TODO: It'd be great to know which one is failing.
            raise ValueError("Failed to construct prover for a conjunct")

        return AndProver(self, subs)

    def get_verifier(self):
        """
        Constructs a Verifier for the and-proof, based on a list of the Verifiers of each subproof.
        """
        return AndVerifier(self, [sub.get_verifier() for sub in self.subproofs])

    def get_randomizers(self):
        """
        Create a dictionary of randomizers by querying the subproofs' maps and merging them.
        """
        random_vals = {}

        # Pair each Secret to one generator. Overwrites when a Secret re-occurs but since the
        # associated bases should yield groups of same order, it's fine.
        dict_name_gen = {s: g for s, g in zip(self.get_secret_vars(), self.get_bases())}

        # Pair each Secret to a randomizer.
        for u in dict_name_gen:
            random_vals[u] = dict_name_gen[u].group.order().random()

        return random_vals

    def prepare_simulate_proof(self):
        for sub in self.subproofs:
            sub.prepare_simulate_proof()

    def simulate_proof(self, responses_dict=None, challenge=None):
        """
        Simulate the And proof

        To do so, draw a global challenge, a global dictionary of responses (for consistency) and
        simulate each subproof.

        Gathers the commitments, and pack everything into a :py:class:`base.SimulationTranscript`.

        Args:
            responses_dict: A dictionary of responses to override (could come from an upper And
                Proof, for example). Draw randomly if None.
            challenge: The challenge to use in the proof. Draw one if None.
        """
        # Fill the missing positions of the responses dictionary
        responses_dict = self.update_randomizers(responses_dict)

        if challenge is None:
            challenge = get_random_num(CHALLENGE_LENGTH)
        com = []
        resp = []
        precom = []

        # Simulate all subproofs and gather their attributes, repack them in a unique
        # SimulationTranscript.
        for sub in self.subproofs:
            simulation = sub.simulate_proof(
                challenge=challenge, responses_dict=responses_dict
            )
            com.append(simulation.commitment)
            resp.append(simulation.responses)
            precom.append(simulation.precommitment)

        return SimulationTranscript(
            commitment=com, challenge=challenge, responses=resp, precommitment=precom
        )

    def validate_secrets_reoccurence(self, forbidden_secrets=None):
        """
        Check re-occuring secrets both inside and outside an or-proof.

        This method gets the list of all secrets in the tree and triggers a depth-first search for
        or-proofs

        Args:
            forbidden_secrets: A list of all the secrets in the mother proof.

        Raises:
            :py:class:`exceptions.InvalidSecretsError`: If any secrets re-occur in an
                unsupported way.
        """
        if forbidden_secrets is None:
            forbidden_secrets = self.get_secret_vars().copy()
        for p in self.subproofs:
            p.validate_secrets_reoccurence(forbidden_secrets)


class AndProver(Prover):
    def __init__(self, proof, subprovers):
        """
        Constructs a Prover for an and-proof, from a list of valid subprovers.
        """
        self.subs = subprovers
        self.stmt = proof

    def precommit(self):
        """
        Computes the precommitment for an and-proof.

        This precommitment is the list of precommitments of the subprovers.

        If not applicable (no subprover outputs a precommitment), returns None.
        """
        precommitment = []
        for index, sub in enumerate(self.subs):
            # Collect precommitments one by one
            sub_precommitment = sub.precommit()
            if sub_precommitment is not None:
                if len(precommitment) == 0:
                    precommitment = [None] * len(self.subs)
                precommitment[index] = sub_precommitment

        # If any precommitment is valid, return the list. If all were None, return None.
        return precommitment if len(precommitment) != 0 else None

    def internal_commit(self, randomizers_dict=None):
        """
        Compute the internal commitment.

        Args:
            randomizers_dict: Mapping from secrets to randomizers.
        """
        # Now that we have constructed the proofs, validate.
        self.stmt.validate_composition()

        randomizers_dict = self.stmt.update_randomizers(randomizers_dict)
        self.commitment = []
        for sub in self.subs:
            self.commitment.append(
                sub.internal_commit(randomizers_dict=randomizers_dict)
            )
        return self.commitment

    def compute_response(self, challenge):
        """
        Return a list of responses of each subprover.
        """
        return [sub.compute_response(challenge) for sub in self.subs]


class AndVerifier(Verifier):
    def __init__(self, proof, subverifiers):
        self.subs = subverifiers
        self.stmt = proof

    def send_challenge(self, commitment, ignore_statement_hash_checks=False):
        """
        Store the received commitment and generate a challenge.

        Additionally checks the received hashed statement matches the one of the current proof. Only
        called at the highest level or in extended proofs.

        Args:
            commitment: A tuple (statement, actual_commitment) with
                actual_commitment a list of commitments, one for each subproof.
            ignore_statement_hash_checks: Optional parameter to deactivate the
                statement check. In this case, the commitment parameter is
                simply the actual commitment. Useful in 2-level proofs for which
                we don't check the inner statements.
        """
        if ignore_statement_hash_checks:
            self.commitment = commitment
        else:
            statement, self.commitment = commitment
            self.stmt.check_statement(statement)
        self.challenge = get_random_num(CHALLENGE_LENGTH)
        return self.challenge

    def check_responses_consistency(self, responses, responses_dict=None):
        """
        Check that the responses are consistent for re-occurring secret names.

        Iterates through the subverifiers, gives them the responses related to
        them and constructs a response dictionary.  If an inconsistency if found
        during this build, returns False.

        Args:
            responses: Rreceived list of responses for each subproof.
            responses_dict: Dictionary to construct and use for comparison.
        """
        if responses_dict is None:
            responses_dict = {}

        for index, sub in enumerate(self.subs):
            if not sub.check_responses_consistency(responses[index], responses_dict):
                return False
        return True

    def process_precommitment(self, precommitment):
        """
        Distribute the list of precommitments to the subverifiers.
        """
        if precommitment is None:
            return
        for index, sub in enumerate(self.subs):
            sub.process_precommitment(precommitment[index])



def get_default_attr(obj, attr, default_value=None):
    """
    Get attribute by name. If does not exist, set it.

    >>> class Klass: pass
    >>> a = Klass()
    >>> hasattr(a, "answer")
    False
    >>> get_default_attr(a, "answer", 42)
    42
    >>> a.answer
    42

    """
    if not hasattr(obj, attr):
        setattr(obj, attr, default_value)
    return getattr(obj, attr)



"""
Wrapper around ``bplib`` points that ensures additive notation for all points.
"""

import attr

from bplib.bp import BpGroup, G1Elem, G2Elem, GTElem

import petlib.pack as pack
import msgpack


class BilinearGroupPair:
    """
    A bilinear group pair.

    Contains two origin groups G1, G2 and the image group GT. The underlying
    ``bplib.bp.BpGroup`` object is also embedded.
    """

    def __init__(self, bp_group=None):
        if bp_group is None:
            self.bpgp = BpGroup()
        self.GT = GTGroup(self)
        self.G1 = G1Group(self)
        self.G2 = G2Group(self)

    def groups(self):
        """
        Returns the three groups in the following order :  G1, G2, GT.
        """
        return self.G1, self.G2, self.GT


class GTGroup:
    """
    Wrapper for the GT group with additive points.

    Allows to retrieve groups G1 and G2.

    The group ID is set to 0 to allow comparisons between groups of different
    types to raise an explicit Exception.

    Args:
        bp (:py:class:`BilinearGroupPair`): Group pair.
    """

    def __init__(self, bp):
        self.bp = bp
        self.gen = None
        self.inf = None

    def infinite(self):
        if self.inf is None:
            self.inf = AdditivePoint(self.generator().pt.one(self.bp.bpgp), self.bp)
        return self.inf

    def order(self):
        return self.bp.bpgp.order()

    def generator(self):
        if self.gen is None:
            self.gen = self.bp.G1.generator().pair(self.bp.G2.generator())
        return self.gen

    def sum(self, points):
        res = self.infinite()
        for p in points:
            res = res + p
        return res

    def wsum(self, weights, generators):
        res = self.infinite()
        for w, g in zip(weights, generators):
            res = res + w * g
        return res


# TODO: Why should this not just be called GTPoint?
class AdditivePoint:
    """
    A wrapper for GT points that uses additive notation.

    Args:
        pt (``bplib.bp.GTElem``): Wrapped point.
        bp (:py:class:`BilinearGroupPair`): Group pair.
    """

    def __init__(self, pt, bp):
        self.pt = pt
        self.bp = bp
        self.group = self.bp.GT

    def export(self, form=0):
        return self.pt.export(form) if form else self.pt.export()

    def __mul__(self, nb):
        """
        Overrides the multiplicative syntax by an additive one.

        Special case in 0 as the underlying ``bplib`` function is broken for
        this value.
        """
        if nb == 0:
            return AdditivePoint(self.pt / self.pt, self.bp)
        return AdditivePoint(self.pt ** nb, self.bp)

    def __eq__(self, other):
        return self.pt == other.pt

    def __add__(self, other):
        """
        Replace the multiplicative syntax between two points by an additive one.
        """
        return AdditivePoint(self.pt * (other.pt), self.bp)

    __rmul__ = __mul__

    def __repr__(self):
        return "GTPt(" + str(self.pt.__hash__()) + ")"


class G1Point:
    """
    Wrapper for G1 points so they can be paired with a G2 point.

    Args:
        pt (``bplib.bp.G1Point``): Point.
        bp (:py:class:`BilinearGroupPair`): Group pair.
    """

    def __init__(self, pt, bp):
        self.pt = pt
        self.bp = bp
        self.group = self.bp.G1

    def __eq__(self, other):
        return self.pt == other.pt

    def __add__(self, other):
        return G1Point(self.pt + other.pt, self.bp)

    def __sub__(self, other):
        return self + (-1 * other)

    def __mul__(self, nb):
        return G1Point(self.pt * nb, self.bp)

    def export(self, form=0):
        return self.pt.export(form) if form else self.pt.export()

    def __eq__(self, other):
        return self.pt == other.pt

    __rmul__ = __mul__

    def pair(self, other):
        return AdditivePoint(self.bp.bpgp.pair(self.pt, other.pt), self.bp)

    def __repr__(self):
        return "G1Pt(" + str(self.pt.__hash__()) + ")"


class G2Point:
    """
    Wrapper for G2 points.

    Args:
        pt (``bplib.bp.G2Point``): Point.
        bp (:py:class:`BilinearGroupPair`): Group pair.
    """

    def __init__(self, pt, bp):
        self.pt = pt
        self.bp = bp
        self.group = self.bp.G2

    def __eq__(self, other):
        return self.pt == other.pt

    def __add__(self, other):
        return G2Point(self.pt + other.pt, self.bp)

    def __sub__(self, other):
        return self + (-1 * other)

    def __mul__(self, nb):
        return G2Point(self.pt * nb, self.bp)

    def export(self, form=0):
        return self.pt.export(form) if form else self.pt.export()

    def __eq__(self, other):
        return self.pt == other.pt

    __rmul__ = __mul__

    def __repr__(self):
        return "G2Pt(" + str(self.pt.__hash__()) + ")"


class G1Group:
    """
    Wrapper for G1 that behaves like normal ``petlib.ec.EcGroup``.

    Args:
        bp (:py:class:`BilinearGroupPair`): Group pair.
    """

    def __init__(self, bp):
        self.bp = bp
        self.gen = None
        self.inf = None

    def generator(self):
        if self.gen is None:
            self.gen = G1Point(self.bp.bpgp.gen1(), self.bp)
        return self.gen

    def infinite(self):
        if self.inf is None:
            self.inf = G1Point(self.generator().pt.inf(self.bp.bpgp), self.bp)
        return self.inf

    def order(self):
        return self.bp.bpgp.order()

    def __eq__(self, other):
        return self.bp.bpgp == other.bp.bpgp and self.__class__ == other.__class__

    def hash_to_point(self, string):
        return G1Point(self.bp.bpgp.hashG1(string), self.bp)

    # TODO throw these on a base class
    def sum(self, points):
        res = self.infinite()
        for p in points:
            res = res + p
        return res

    # TODO throw these on a base class
    def wsum(self, weights, generators):
        res = self.infinite()
        for w, g in zip(weights, generators):
            res = res + w * g
        return res


class G2Group:
    """
    Wrapper for the G2 group.

    Args:
        bp (:py:class:`BilinearGroupPair`): Group pair.

    """

    def __init__(self, bp):
        self.bp = bp
        self.gen = None
        self.inf = None

    def generator(self):
        if self.gen is None:
            self.gen = G2Point(self.bp.bpgp.gen2(), self.bp)
        return self.gen

    def infinite(self):
        if self.inf is None:
            self.inf = G2Point(self.generator().pt.inf(self.bp.bpgp), self.bp)
        return self.inf

    def order(self):
        return self.bp.bpgp.order()

    # TODO throw these on a base class
    def sum(self, points):
        res = self.infinite()
        for p in points:
            res = res + p
        return res

    def wsum(self, weights, generators):
        res = self.infinite()
        for w, g in zip(weights, generators):
            res = res + w * g
        return res


def pt_enc(obj):
    """Encoder for the wrapped points."""
    nid = obj.bp.bpgp.nid
    data = obj.pt.export()
    packed_data = msgpack.packb((nid, data))
    return packed_data


def pt_dec(bptype, xtype):
    """Decoder for the wrapped points."""

    def dec(data):
        nid, data = msgpack.unpackb(data)
        bp = BilinearGroupPair()
        pt = bptype.from_bytes(data, bp.bpgp)
        return xtype(pt, bp)

    return dec


# Register encoders and decoders for pairing points
pack.register_coders(G1Point, 111, pt_enc, pt_dec(G1Elem, G1Point))
pack.register_coders(G2Point, 112, pt_enc, pt_dec(G2Elem, G2Point))
pack.register_coders(AdditivePoint, 113, pt_enc, pt_dec(GTElem, AdditivePoint))



r"""
Range proof: ZK proof that a committed value lies within a range.

.. math::

    PK \{ (r, x): \underbrace{C = x G + r H}_{Commitment} \land \underbrace{l \leq x < u}_{Range} \}

This module implements a Schoenmakers' range proof, a conjuction of or-proofs for each bit of the
value.

"""

import warnings

from petlib.bn import Bn
from petlib.ec import EcGroup

from zksk import Secret
from zksk.primitives.dlrep import DLRep
from zksk.exceptions import ValidationError
from zksk.extended import ExtendedProofStmt
from zksk.utils import make_generators, get_random_num, ensure_bn
from zksk.composition import AndProofStmt


def decompose_into_n_bits(value, n):
    """Array of bits, least significant bit first"""
    if value < 0:
        raise Exception("Can't represent negative values")

    base = [1 if value.is_bit_set(b) else 0 for b in range(value.num_bits())]

    extra_bits = n - len(base)
    if extra_bits < 0:
        raise Exception("Not enough bits to represent value")

    return base + [0] * extra_bits


class PowerTwoRangeStmt(ExtendedProofStmt):
    r"""
    A power-two range proof statement.

    .. math::

        PK \{ (r, x): C = x G + r H \land 0 \leq x < 2^n \}

    Args:
        com: Value of the Pedersen commitment, :math:`C = x G + r H`
        g: First commitment base point :math:`G`
        h: Second commitment base point :math:`H`
        num_bits: The number of bits of the committed value :math:`n`
        x: Value for which we construct a range proof (prover only)
        randomizer: Randomizer of the commitment :math:`r` (prover only)
    """

    def __init__(self, com, g, h, num_bits, x=None, randomizer=None):
        if not x.value is None and not randomizer.value is None:
            self.x = x
            self.randomizer = randomizer
            self.is_prover = True

            # Ensure secret is in range
            self.x.value = ensure_bn(self.x.value)
            if self.x.value < 0:
                warnings.warn("Secret is negative")
            if self.x.value.num_bits() > num_bits:
                warnings.warn("Secret has more than {} bits".format(num_bits))
        else:
            self.is_prover = False

        # TODO: Should we combine com with the inner proof?
        self.com = com
        self.g = g
        self.h = h
        self.order = g.group.order()
        self.num_bits = num_bits

        # The constructed proofs need extra randomizers as secrets
        self.randomizers = [Secret() for _ in range(self.num_bits)]

    def precommit(self):
        """
        Commit to the bit-decomposition of the value.
        """
        actual_value = ensure_bn(self.x.value)
        value_as_bits = decompose_into_n_bits(actual_value, self.num_bits)

        # Set true value to computed secrets
        for rand in self.randomizers:
            rand.value = self.order.random()

        precommitment = {}
        precommitment["Cs"] = [
            b * self.g + r.value * self.h
            for b, r in zip(value_as_bits, self.randomizers)
        ]

        # Compute revealed randomizer
        rand = Bn(0)
        power = Bn(1)
        for r in self.randomizers:
            rand = rand.mod_add(r.value * power, self.order)
            power *= 2
        rand = rand.mod_sub(self.randomizer.value, self.order)
        precommitment["rand"] = rand

        return precommitment

    def construct_stmt(self, precommitment):
        """
        Construct the internal proof statement.
        """
        if self.is_prover:
            # Indicators that tell us which or-clause is true
            actual_value = ensure_bn(self.x.value)
            value_as_bits = decompose_into_n_bits(actual_value, self.num_bits)
            zero_simulated = [b == 1 for b in value_as_bits]
            one_simulated = [b == 0 for b in value_as_bits]

        bit_proofs = []
        for i in range(self.num_bits):
            p0 = DLRep(precommitment["Cs"][i], self.randomizers[i] * self.h)
            p1 = DLRep(precommitment["Cs"][i] - self.g, self.randomizers[i] * self.h)

            # When we are a prover, mark which disjunct is true
            if self.is_prover:
                p0.set_simulated(zero_simulated[i])
                p1.set_simulated(one_simulated[i])

            bit_proofs.append(p0 | p1)

        return AndProofStmt(*bit_proofs)

    def simulate_precommit(self):
        randomizers = [self.order.random() for _ in range(self.num_bits)]
        precommitment = {}
        precommitment["Cs"] = [r * self.h for r in randomizers]
        precommitment["Cs"][0] += self.com

        # Compute revealed randomizer
        rand = Bn(0)
        power = Bn(1)
        for r in randomizers:
            rand = rand.mod_add(r * power, self.order)
            power *= 2
        precommitment["rand"] = rand

        return precommitment

    def validate(self, precommitment):
        """
        Check the commitment to the bit-decomposition is correct.
        """
        rand = precommitment["rand"]

        # Combine bit commitments into value commitment
        combined = self.g.group.infinite()
        power = Bn(1)
        for c in precommitment["Cs"]:
            combined += power * c
            power *= 2

        if combined != self.com + rand * self.h:
            raise ValidationError("The commitments do not combine correctly")


class GenericRangeStmtMaker:
    r"""
    Auxiliary builder class for generic range proofs.

    .. math::

        PK \{ (r, x): x G + r H \land a \leq x < b \}

    See "`Efficient Protocols for Set Membership and Range Proofs`_" by Camenisch
    et al., 2008.

    In practice, use the :py:obj:`zksk.primitives.rangeproof.RangeStmt` object directly:

    >>> group = EcGroup()
    >>> x = Secret(value=3)
    >>> randomizer = Secret(value=group.order().random())
    >>> g = group.hash_to_point(b"1")
    >>> h = group.hash_to_point(b"2")
    >>> lo = 0
    >>> hi = 5
    >>> com = x * g + randomizer * h
    >>> stmt = RangeStmt(com.eval(), g, h, lo, hi, x, randomizer)
    >>> nizk = stmt.prove()
    >>> stmt.verify(nizk)
    True

    See :py:meth:`GenericRangeStmtMaker.__call__` for the construction signature.

    .. `Efficient Protocols for Set Membership and Range Proofs`:
        https://infoscience.epfl.ch/record/128718/files/CCS08.pdf

    """

    def __call__(self, com, g, h, a, b, x, r):
        """
        Get a conjunction of two range-power-of-two proofs.

        Args:
            com: Value of the Pedersen commitment, :math:`C = x G + r H`
            g: First commitment base point :math:`G`
            h: Second commitment base point :math:`H`
            a: Lower limit :math:`a`
            b: Upper limit :math:`b`
            x: Value for which we construct a range proof
            r: Randomizer of the commitment :math:`r`
        """
        a = ensure_bn(a)
        b = ensure_bn(b)
        num_bits = (b - a - 1).num_bits()
        offset = Bn(2) ** num_bits - (b - a)

        com_shifted1 = com - a * g
        com_shifted2 = com_shifted1 + offset * g
        x1 = Secret()
        x2 = Secret()
        if x.value is not None:
            x1.value = x.value - a
            x2.value = x.value - a + offset

            # Ensure secret is in range
            if x.value < a or x.value >= b:
                warnings.warn("Secret outside of given range [{}, {})".format(a, b))

        com_stmt = DLRep(com, x * g + r * h)

        p1 = PowerTwoRangeStmt(
            com=com_shifted1, g=g, h=h, num_bits=num_bits, x=x1, randomizer=r,
        )

        p2 = PowerTwoRangeStmt(
            com=com_shifted2, g=g, h=h, num_bits=num_bits, x=x2, randomizer=r,
        )

        return com_stmt & p1 & p2


class GenericRangeOnlyStmtMaker:
    r"""
    Auxiliary builder class for generic range proofs.

    .. math::
        PK \{ (x): a \leq x < b \}

    See "`Efficient Protocols for Set Membership and Range Proofs`_" by Camenisch
    et al., 2008.

    .. _`Efficient Protocols for Set Membership and Range Proofs`:
        https://infoscience.epfl.ch/record/128718/files/CCS08.pdf

    In practice, use the :py:obj:`zksk.primitives.rangeproof.RangeStmt` object directly:

    >>> x = Secret(value=3)
    >>> lo = 0
    >>> hi = 5
    >>> stmt = RangeOnlyStmt(lo, hi, x)
    >>> nizk = stmt.prove()
    >>> stmt.verify(nizk)
    True

    See :py:meth:`GenericRangeStmtMaker.__call__` for the construction signature.
    """

    def __call__(self, a, b, x=None):
        """
        Get a conjunction of two range-power-of-two proofs.
        Args:
            a: Lower limit :math:`a`
            b: Upper limit :math:`b`
            x: Value for which we construct a range proof
        """
        group = EcGroup()
        g = group.hash_to_point(b"g")
        h = group.hash_to_point(b"h")

        r = Secret(value=group.order().random())
        com = (x * g + r * h).eval()

        a = ensure_bn(a)
        b = ensure_bn(b)
        num_bits = (b - a - 1).num_bits()
        offset = Bn(2) ** num_bits - (b - a)
        com_shifted1 = com - a * g
        com_shifted2 = com_shifted1 + offset * g

        x1 = Secret()
        x2 = Secret()
        if x is not None:
            x1.value = x.value - a
            x2.value = x.value - a + offset

        com_stmt = DLRep(com, x * g + r * h)
        p1 = PowerTwoRangeStmt(
            com=com_shifted1, g=g, h=h, num_bits=num_bits, x=x1, randomizer=r,
        )
        p2 = PowerTwoRangeStmt(
            com=com_shifted2, g=g, h=h, num_bits=num_bits, x=x2, randomizer=r,
        )

        return com_stmt & p1 & p2


# TODO: Make a regular class.
RangeStmt = GenericRangeStmtMaker()
RangeOnlyStmt = GenericRangeOnlyStmtMaker()



r"""
ZK proof of inequality of two discrete logarithms.

.. math::

    PK\{ (x): H_0 = x h_0 \land H_1 \neq x h_1 \}

See Protocol 1 in "`Thinking Inside the BLAC Box: Smarter Protocols for Faster Anonymous
Blacklisting`_" by Henry and Goldberg, 2013:

.. _`Thinking Inside the BLAC Box: Smarter Protocols for Faster Anonymous
    Blacklisting`: https://www.cypherpunks.ca/~iang/pubs/blacronym-wpes.pdf
"""

from zksk.expr import Secret, wsum_secrets
from zksk.exceptions import ValidationError
from zksk.extended import ExtendedProofStmt, ExtendedVerifier
from zksk.composition import AndProofStmt
from zksk.primitives.dlrep import DLRep


class DLNotEqual(ExtendedProofStmt):
    r"""
    ZK proof statement of inequality of two discrete logarithms.

    .. math::

        PK\{ (x): H_0 = x h_0 \land H_1 \neq x h_1 \}

    The statement is constructed from two pairs: :math:`(H_0, h_0)`, :math:`(H_1, h_1)`, and a
    :py:class:`expr.Secret` object representing a secret :math:`x`.

    The proof can be made `binding`: bind the :math:`x` to another proof. If the proof is not
    binding, it is not possible to assert that the same :math:`x` was used in any other proof (even
    in, say, an AND conjunction).

    Args:
        valid_pair (tuple): Pair of two Elliptic curve points :math:`(H_0, h_0)` such that
            :math:`H_0 = x h_0`
        invalid_pair (tuple): Pair of two Elliptic curve points :math:`(H_1, h_1)` such that
            :math:`H_1 \neq x h_1`
        x (:py:class:`expr.Secret`): Secret.
        bind (bool): Whether the proof is binding.
        simulated (bool): If this proof is a part of an or-proof: whether it should be simulated.
    """

    def __init__(self, valid_pair, invalid_pair, x, bind=False, simulated=False):
        if len(valid_pair) != 2 or len(invalid_pair) != 2:
            raise TypeException("The valid_pair and invalid_pair must be pairs")

        self.x = x

        # The internal ZK proof uses two constructed secrets
        self.alpha, self.beta = Secret(), Secret()

        self.lhs = [valid_pair[0], invalid_pair[0]]
        self.g = valid_pair[1]
        self.h = invalid_pair[1]

        self.bind = bind
        self.set_simulated(simulated)

    def precommit(self):
        """Build the left-hand side of the internal proof statement."""
        order = self.g.group.order()
        blinder = order.random()

        # Set the value of the two internal secrets
        self.alpha.value = self.x.value * blinder % order
        self.beta.value = -blinder % order

        precommitment = blinder * (self.x.value * self.h - self.lhs[1])
        return precommitment

    def construct_stmt(self, precommitment):
        """
        Build the internal proof statement.

        See the formula in Protocol 1 of the `Thinking Inside the BLAC Box: Smarter Protocols for
        Faster Anonymous Blacklisting` paper.
        """
        infty = self.g.group.infinite()
        p1 = DLRep(infty, self.alpha * self.g + self.beta * self.lhs[0])
        p2 = DLRep(precommitment, self.alpha * self.h + self.beta * self.lhs[1])
        statements = [p1, p2]

        if self.bind:
            # If the binding parameter is set, we add a DLRep member repeating
            # the first member without randomizing the secret.
            statements.append(DLRep(self.lhs[0], self.x * self.g))

        return AndProofStmt(*statements)

    def validate(self, precommitment):
        """
        Verify the the proof statement is indeed proving the inequality of discret logs.
        """
        if precommitment == self.g.group.infinite():
            raise ValidationError("The commitment should not be the unity element")

    def simulate_precommit(self):
        """
        Draw a base at random (not unity) from the bases' group.
        """
        group = self.g.group
        precommitment = group.order().random() * group.generator()
        return precommitment



import random

import pytest

from petlib.bn import Bn
from petlib.ec import EcGroup

from zksk import DLRep, Secret
from zksk.exceptions import (
    InvalidExpression,
    InvalidSecretsError,
    ValidationError,
    GroupMismatchError,
)
from zksk.composition import AndProofStmt, OrProofStmt
from zksk.expr import wsum_secrets
from zksk.utils import make_generators
from zksk.base import NIZK

@pytest.fixture
def params(group):
    n1 = 3
    n2 = 4
    generators1 = make_generators(n1)
    generators2 = make_generators(n2)
    x0 = Secret()
    x1 = Secret()
    x2 = Secret()
    x3 = Secret()
    x4 = Secret()
    x5 = Secret()
    secrets = [x0, x1, x2, x3, x4, x5]
    secrets_dict = dict(
        [
            (x0, Bn(1)),
            (x1, Bn(2)),
            (x2, Bn(5)),
            (x3, Bn(100)),
            (x4, Bn(43)),
            (x5, Bn(10)),
        ]
    )

    sum_1 = group.wsum(
        [secrets_dict[x0], secrets_dict[x1], secrets_dict[x2]], generators1
    )
    secrets_2 = [secrets_dict[x0]]
    for i in range(3, 6):
        secrets_2.append(secrets_dict[secrets[i]])

    sum_2 = group.wsum(secrets_2, generators2)
    p1 = DLRep(sum_1, wsum_secrets([x0, x1, x2], generators1))

    p2 = DLRep(sum_2, wsum_secrets([x0, x3, x4, x5], generators2))
    return p1, p2, secrets_dict


def verify(verifier, prover):
    commitment = prover.commit()
    challenge = verifier.send_challenge(commitment)
    response = prover.compute_response(challenge)
    return verifier.verify(response)


def verify_proof(proof, secrets):
    prov = proof.get_prover(secrets)
    verif = proof.get_verifier()
    com = prov.commit()
    chal = verif.send_challenge(com)
    resp = prov.compute_response(chal)
    return verif.verify(resp)


def test_and_proof_same_environment(params):
    p1, p2, secrets_dict = params
    and_proof = AndProofStmt(p1, p2)

    prover = and_proof.get_prover(secrets_dict)
    verifier = and_proof.get_verifier()
    assert verify(verifier, prover)


def test_and_proof_different_environments(params):
    x, y = Secret(), Secret()
    p1, p2, secrets_dict = params
    and_proof = AndProofStmt(p1, p2)
    prover = and_proof.get_prover(secrets_dict)
    verifier = and_proof.get_verifier()
    assert verify(verifier, prover)


def test_and_proof_partially_defined_secrets():
    generators = make_generators(2)
    x = Secret(value=4)
    x2 = Secret()
    p1 = DLRep(4 * generators[0], x * generators[0])
    p2 = DLRep(3 * generators[1], x2 * generators[1])
    andp = p1 & p2
    tr = andp.prove({x2: 3})
    assert andp.verify(tr)


def test_and_proof_fails_when_bases_belong_to_different_groups(group):
    """
    An alien EcPt is inserted in the generators
    """
    g1 = group.generator()
    other_group = EcGroup(706)
    assert group != other_group
    g2 = other_group.generator()

    x = Secret(value=Bn(42))
    y1 = group.wsum([x.value], [g1])
    y2 = other_group.wsum([x.value], [g2])

    p1 = DLRep(y1, wsum_secrets([x], [g1]))
    p2 = DLRep(y2, wsum_secrets([x], [g2]))

    and_proof = AndProofStmt(p1, p2)
    prover = and_proof.get_prover()
    verifier = and_proof.get_verifier()

    # An exception should be raised because of a shared secrets linked to two different groups
    with pytest.raises(GroupMismatchError):
        verify(verifier, prover)


def test_and_proof_fails_when_secret_is_wrong(params, group):
    p1, p2, secrets_dict = params
    and_proof = AndProofStmt(p1, p2)
    sec = secrets_dict.copy()
    u = list(sec.keys())
    sec[u[0]] = group.order().random()
    prover = and_proof.get_prover(sec)
    verifier = and_proof.get_verifier()
    assert not verify(verifier, prover)


def test_multiple_and_proofs(params):
    p1, p2, secrets_dict = params
    and_proof = AndProofStmt(p1, p2, p2, p1, p1, p1, p2)
    prover = and_proof.get_prover(secrets_dict)
    verifier = and_proof.get_verifier()
    assert verify(verifier, prover)


def test_compose_and_proofs_1(params):
    p1, p2, secrets_dict = params
    p3 = AndProofStmt(p1, p2)
    p4 = AndProofStmt(AndProofStmt(p1, p2), p1)
    prover = p4.get_prover(secrets_dict)
    verifier = p4.get_verifier()
    assert verify(verifier, prover)


def test_compose_and_proofs_2(params):
    p1, p2, secrets_dict = params
    p3 = AndProofStmt(p1, p2)
    p = AndProofStmt(AndProofStmt(p1, AndProofStmt(p3, AndProofStmt(p1, p2))), p2)
    prover = p.get_prover(secrets_dict)
    verifier = p.get_verifier()
    assert verify(verifier, prover)


def test_and_proof_simulation_1(group):
    n = 3
    secret_values = [Bn(i) for i in range(n)]
    secrets = [Secret() for _ in range(n)]
    generators = make_generators(n, group)
    lhs = group.wsum(secret_values, generators)

    subproof1 = DLRep(lhs, wsum_secrets(secrets, generators))
    subproof2 = DLRep(lhs, wsum_secrets(secrets, generators))
    andp = AndProofStmt(subproof1, subproof2)
    andv = andp.get_verifier()
    tr = andp.simulate_proof()
    tr.stmt_hash = andp.prehash_statement().digest()
    assert not andv.verify_nizk(tr)


def test_and_proof_simulation_2(group):
    n = 3
    secret_values = [Bn(i) for i in range(n)]
    secrets = [Secret() for _ in range(n)]
    generators = make_generators(n, group)
    lhs = group.wsum(secret_values, generators)

    subproof1 = DLRep(lhs, wsum_secrets(secrets, generators))
    subproof2 = DLRep(lhs, wsum_secrets(secrets, generators))
    andp = AndProofStmt(subproof1, subproof2)
    tr = andp.simulate()
    assert andp.verify_simulation_consistency(tr)
    assert not andp.verify(tr)


def test_and_proof_non_interactive(params):
    p1, p2, secrets = params
    p = AndProofStmt(p1, p2)
    message = "whatever"
    tr = p.prove(secrets, message=message)
    assert p.verify(tr, message=message)


def test_and_proof_non_interactive_fails_when_wrong_secrets(params, group):
    p1, p2, secrets = params
    andp = AndProofStmt(p1, p2)

    bad_secrets = secrets.copy()
    u = list(bad_secrets.keys())
    bad_secrets[u[0]] = group.order().random()

    message = "whatever"
    tr = andp.prove(bad_secrets, message=message)
    assert not andp.verify(tr, message=message)


def test_and_proof_infix_operator(params):
    p1, p2, secrets_dict = params
    and_proof = p1 & p2 & p1
    prover = and_proof.get_prover(secrets_dict)
    verifier = and_proof.get_verifier()
    assert verify(verifier, prover)


def test_and_proof_with_complex_expression(group):
    g = group.generator()
    g1 = 2 * g
    g2 = 5 * g
    g3 = 10 * g
    x1 = Secret()
    x2 = Secret()
    x3 = Secret()
    proof = DLRep(10 * g1 + 15 * g2, x1 * g1 + x2 * g2) & DLRep(
        15 * g1 + 35 * g3, x2 * g1 + x3 * g3
    )
    prover = proof.get_prover({x1: 10, x2: 15, x3: 35})
    verifier = proof.get_verifier()
    assert verify(verifier, prover)


def test_or_proof(params):
    p1, p2, secrets = params
    orproof = OrProofStmt(p1, p2, p1, p2, p1, p2)
    prov = orproof.get_prover(secrets)
    verif = orproof.get_verifier()
    com = prov.commit()
    chal = verif.send_challenge(com)
    resp = prov.compute_response(chal)

    # Here we see that some responses have an identical first element.
    # The only one with a different first element is the non-simulated one.
    assert verif.verify(resp)


def test_or_proof_manual(params):
    """
    TODO: Clarify what is being tested here.
    """
    p1, p2, secrets = params
    orproof = OrProofStmt(p1, p2, p1, p2, p1, p2)

    subproofs = orproof.subproofs
    rep = 0
    chosen = []
    sims = [True]
    while rep < 10:
        sims = []

        # Make random subproofs simulated.
        for proof in subproofs:
            is_simulated = random.choice([True, False])
            sims.append(is_simulated)
            proof.set_simulated(is_simulated)

        if all(sims):
            continue

        for i in range(30):
            # Choose a subproof, look if it was a valid choice, store the result
            prov = orproof.get_prover(secrets)
            chosen.append(sims[orproof.chosen_idx] == False)

        rep += 1

    assert all(chosen)


def test_and_or_proof_composition(params):
    p1, p2, secrets = params
    g1 = 7 * p1.bases[0]
    g2 = 8 * p1.bases[0]
    xb = Secret(name="xb")
    xa = Secret(name="xa")

    p0 = DLRep(7 * g1 + 18 * g2, xb * g1 + xa * g2)
    secrets[xb] = 7
    secrets[xa] = 18

    orproof = OrProofStmt(p1, p2)
    andp = AndProofStmt(orproof, p0)
    andp = AndProofStmt(andp, DLRep(15 * p1.bases[0], Secret(value=15) * p1.bases[0]))

    prover = andp.get_prover(secrets)
    verifier = andp.get_verifier()
    assert verify(verifier, prover)


def test_or_and_proof_composition(params):
    p1, p2, secrets = params
    andp = AndProofStmt(p1, p2)

    g1 = 7 * p1.bases[0]
    g2 = 8 * p1.bases[0]
    xb = Secret(name="xb")
    xa = Secret(name="xa")
    p0 = DLRep(7 * g1 + 18 * g2, xb * g1 + xa * g2)
    secrets[xa] = 7
    secrets[Secret(name="xc")] = 18

    orproof = OrProofStmt(p0, andp)
    prover = orproof.get_prover(secrets)
    verifier = orproof.get_verifier()
    assert verify(verifier, prover)


def test_or_or_proof_composition(params):
    p1, p2, secrets = params

    first_or = OrProofStmt(p1, p2)
    g1 = 7 * p1.bases[0]
    g2 = 8 * p1.bases[0]
    xb = Secret(name="xb")
    xa = Secret(name="xa")

    p0 = DLRep(7 * g1 + 18 * g2, xb * g1 + xa * g2)
    secrets[xa] = 7
    secrets[Secret()] = 18

    orproof = OrProofStmt(p0, first_or)
    prover = orproof.get_prover(secrets)
    verifier = orproof.get_verifier()
    assert verify(verifier, prover)


def test_or_proof_simulation(params):
    p1, p2, secrets = params
    first_or = OrProofStmt(p1, p2)
    tr = first_or.simulate()
    assert first_or.verify_simulation_consistency(tr)
    assert not first_or.verify(tr)


def test_multiple_or_proofs(group, params):
    p1, p2, secrets = params
    g = group.generator()
    x10 = Secret()
    secrets.update({x10: 13})
    orproof = OrProofStmt(p1, OrProofStmt(p2, DLRep(13 * g, x10 * g)))
    assert verify_proof(orproof, secrets)


def test_multiple_or_proofs_composition(group, params):
    p1, p2, secrets = params
    g = group.generator()
    x10 = Secret()
    secrets.update({x10: 13})
    orp1 = OrProofStmt(p2, p1)
    orp2 = OrProofStmt(p1, DLRep(13 * g, x10 * g))
    orproof = OrProofStmt(orp1, p2, orp2)
    assert verify_proof(orproof, secrets)


def test_or_proof_infix_operator(params):
    p1, p2, secrets = params
    orproof = p1 | p2
    assert verify_proof(orproof, secrets)


def test_multiple_or_proof_infix_operator(group, params):
    p1, p2, secrets = params
    g = group.generator()
    x10 = Secret()
    secrets.update({x10: 13})
    orproof = p1 | p2 | DLRep(13 * g, x10 * g)
    assert verify_proof(orproof, secrets)


def test_or_non_interactive(params):
    p1, p2, secrets = params
    p = OrProofStmt(p1, p2)
    message = "whatever"
    tr = p.prove(secrets, message=message)
    assert p.verify(tr, message=message)


def test_or_non_interactive_serialization(params):
    p1, p2, secrets = params
    p = OrProofStmt(p1, p2)
    message = "whatever"
    tr = p.prove(secrets, message=message)
    tr_enc = tr.serialize()
    tr_dec = NIZK.deserialize(tr_enc)
    assert tr_dec.challenge == tr.challenge
    assert len(tr_dec.responses) == len(tr.responses)
    for resp1, resp2 in zip(tr_dec.responses, tr.responses):
        assert resp1 == resp2
    assert tr_dec.precommitment == tr.precommitment
    assert tr_dec.stmt_hash == tr.stmt_hash
    assert p.verify(tr_dec, message=message)


def test_or_non_interactive_fails_on_wrong_secrets(group, params):
    p1, p2, secrets = params
    p = OrProofStmt(p1, p2)
    bad_secrets = secrets.copy()
    u = list(bad_secrets.keys())
    bad_secrets[u[0]] = group.order().random()
    bad_secrets[u[1]] = group.order().random()
    bad_secrets[u[2]] = group.order().random()
    bad_secrets[u[3]] = group.order().random()

    message = "whatever"
    tr = p.prove(bad_secrets, message=message)
    assert not p.verify(tr, message=message)


def test_malicious_and_proofs():
    x0 = Secret()
    x2 = Secret()
    x1 = Secret()
    generators = make_generators(3)
    g1 = generators[0]
    g2 = generators[1]
    g3 = generators[2]
    secret_dict = {x0: 3, x2: 50, x1: 12}
    mal_secret_dict = {x0: 3, x2: 51}
    andp = AndProofStmt(
        DLRep(12 * g1 + 50 * g2, x1 * g1 + x2 * g2),
        DLRep(3 * g3 + 51 * g2, x0 * g1 + x2 * g2),
    )

    prov = andp.get_prover(secret_dict)
    prov.subs[1].secret_values = mal_secret_dict
    verif = andp.get_verifier()

    com = prov.commit()
    chal = verif.send_challenge(com)
    resp = prov.compute_response(chal)
    with pytest.raises(ValidationError):
        verif.verify(resp)


# Secret used both inside and outside of or clause
def test_invalid_or_composition():
    r = Secret(10)
    g1, g2, g3 = make_generators(3)
    st1 = DLRep(10 * g1, r * g1)

    st21 = DLRep(10 * g2, r * g2)
    st22 = DLRep(12 * g3, r * g3)
    st22.set_simulated()
    st2 = st21 | st22
    st = st1 & st2

    with pytest.raises(InvalidSecretsError):
        st.prove()


# Secret used both inside two different or clauses
def test_invalid_or_composition_inside_two_or():
    r = Secret(10)
    g1, g2, g3, g4 = make_generators(4)
    st11 = DLRep(r.value * g1, r * g1)
    st12 = DLRep(2 * g2, r * g2)
    st12.set_simulated()
    st1 = st11 | st12

    st21 = DLRep(7 * g3, r * g3)
    st21.simluation = True
    st22 = DLRep(r.value * g4, r * g4)
    st2 = st21 | st22
    st = st1 & st2

    with pytest.raises(InvalidSecretsError):
        st.prove()



import re
import pytest

from petlib.ec import EcPt

from zksk.expr import Secret, Expression, wsum_secrets
from zksk.utils import make_generators
from zksk.exceptions import IncompleteValuesError, InvalidExpression


def test_secret_named():
    x = Secret(name="x")
    assert x.name == "x"


def test_secret_same_names():
    x, y = Secret(name="same"), Secret(name="same")
    assert x.name == y.name
    assert hash(x) == hash(y)
    assert x == y


def test_secret_same_names_different_values():
    x, y = Secret(name="same"), Secret(name="same", value=100)
    assert x.name == y.name
    assert hash(x) == hash(y)
    assert x != y


def test_secret_different_anonymous_secrets():
    x, y = Secret(), Secret()
    assert x.name != y.name
    assert hash(x) != hash(y)
    assert x != y


@pytest.mark.parametrize(
    "x,pattern",
    [
        (Secret(), r"Secret\(name='.+'\)"),
        (Secret(name="x"), r"Secret\(name='x'\)"),
        (Secret(value=42), r"Secret\(42, '.+'\)"),
        (Secret(value=42, name="x"), r"Secret\(42, 'x'\)"),
    ],
)
def test_secret_repr(x, pattern):
    assert re.match(pattern, repr(x)) is not None


def test_expr_1(group):
    g = group.generator()
    x = Secret()
    expr = Expression(x, g)
    assert expr.bases == (g,)
    assert expr.secrets == (x,)


def test_expr_2(group):
    g = group.generator()
    x = Secret()
    expr = x * g
    assert expr.bases == (g,)
    assert expr.secrets == (x,)


def test_expr_eval_no_values_specified(group):
    g = group.generator()
    g1 = 2 * g
    g2 = 5 * g

    rhs = Secret(name="x1") * g1 + Secret(name="x2") * g2
    with pytest.raises(IncompleteValuesError):
        rhs.eval()


@pytest.mark.parametrize("other", [Secret(), 2, "gibberish"])
def test_expr_invalid_expression(group, other):
    g = group.generator()
    expr = Secret() * g

    with pytest.raises(InvalidExpression):
        expr + other


@pytest.mark.parametrize("num", [2, 10])
def test_expr_repr(group, num):
    secrets = [Secret() for _ in range(num)]
    generators = make_generators(num, group)
    expr = wsum_secrets(secrets, generators)

    expected_repr = " + ".join(
        "Expression({}, {})".format(x, g) for x, g in zip(secrets, generators)
    )
    assert expected_repr == repr(expr)



#!/usr/bin/env python
import os
import re

from setuptools import setup, find_packages


INSTALL_REQUIRES = ["petlib-fork", "bplib-fork"]
SETUP_REQUIRES = ["pytest-runner"]
TEST_REQUIRES = ["pytest"]
DOC_REQUIRES = ["sphinx", "sphinx_rtd_theme", "m2r"]
DEV_REQUIRES = TEST_REQUIRES + DOC_REQUIRES + \
    ["black", "pre-commit", "pytest-cov"]


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst")) as f:
    long_description = f.read()


with open(os.path.join(here, "zksk/__init__.py")) as f:
    matches = re.findall(r"(__.+__) = \"(.*)\"", f.read())
    for var_name, var_value in matches:
        globals()[var_name] = var_value


setup(
    name="zksk-fork",
    version=__version__,
    description=__description__,
    long_description=long_description,
    author=__author__,
    author_email=__email__,
    packages=find_packages(exclude=["tests"]),
    license=__license__,
    url=__url__,
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    tests_require=TEST_REQUIRES,
    extras_require={"dev": DEV_REQUIRES,
                    "test": TEST_REQUIRES, "doc": DOC_REQUIRES, },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Security :: Cryptography",
        "License :: OSI Approved :: MIT License",
    ],
)



"""
Tiny expression language suitable for expressing statements on discrete logarithm representations

>>> from petlib.ec import EcGroup
>>> group = EcGroup()
>>> g = group.hash_to_point(b"1")
>>> h = group.hash_to_point(b"2")
>>> a = Secret()
>>> b = Secret()
>>> expr = a * g + b * h

"""

import struct
import hashlib

from zksk.exceptions import InvalidExpression, IncompleteValuesError


class Expression:
    """
    Arithmetic expression of secrets and group elements.

    It is an abstraction for :math:`x_0 g_0 + x_1 g_2 + ... + x_n g_n`, where :math:`x_i`-s are
    declared secrets.

    Implementation-wise, parses the sum into an ordered list of Secrets and an ordered list of
    generators.

    Args:
        secret (Secret): Secret object.
        base: Base point on an elliptic curve.
    """

    def __init__(self, secret, base):
        if not isinstance(secret, Secret):
            raise InvalidExpression(
                "In {0} * {1}, the first parameter should be a Secret".format(
                    secret, base
                )
            )
        self._secrets = [secret]
        self._bases = [base]

    def __add__(self, other):
        """
        Merge Expression objects along addition.

        Args:
            other (Expression): Another expression

        Returns:
            Expression: New expression
        """
        if not isinstance(other, Expression):
            raise InvalidExpression(
                "Invalid expression. Only linear combinations of group elements are supported."
            )
        self._secrets.extend(other._secrets)
        self._bases.extend(other._bases)
        return self

    @property
    def secrets(self):
        return tuple(self._secrets)

    @property
    def bases(self):
        return tuple(self._bases)

    def eval(self):
        """Evaluate the expression, if all secret values are available."""
        # TODO: Take secret_dict as optional input.
        for secret in self._secrets:
            if secret.value is None:
                raise IncompleteValuesError(
                    "Secret {0} does not have a value".format(secret.name)
                )

        def ith_mul(i):
            return self._secrets[i].value * self._bases[i]

        summation = ith_mul(0)
        for i in range(1, len(self._secrets)):
            summation += ith_mul(i)
        return summation

    def __repr__(self):
        fragments = []
        for secret, base in zip(self._secrets, self._bases):
            fragments.append("Expression({}, {})".format(secret, base))
        return " + ".join(fragments)


class Secret:
    """
    A secret value in a zero-knowledge proof.

    Args
        name: String to enforce as name of the Secret. Useful for debugging.
        value: Optional secret value.
    """

    # Number of bytes in a randomly-generated name of a secret.
    NUM_NAME_BYTES = 8

    def __init__(self, value=None, name=None):
        if name is None:
            name = self._generate_unique_name()
        self.name = name
        self.value = value

    def _generate_unique_name(self):
        h = struct.pack(">q", super().__hash__())
        return hashlib.sha256(h).hexdigest()[: self.NUM_NAME_BYTES * 4]

    def __mul__(self, base):
        """
        Construct an expression that represents this secrets multiplied by the base.

        Args:
            base: Base point on an elliptic curve.

        Returns:
            Expression: Expression that corresponds to :math:`x G`
        """
        return Expression(self, base)

    __rmul__ = __mul__

    def __repr__(self):
        if self.value is None and self.name is None:
            return "Secret()"
        elif self.value is None:
            return "Secret(name={})".format(repr(self.name))
        elif self.name is None:
            return "Secret({})".format(self.value)
        else:
            return "Secret({}, {})".format(self.value, repr(self.name))

    def __hash__(self):
        return hash(("Secret", self.name))

    def __eq__(self, other):
        return (hash(self) == hash(other)) and self.value == other.value


def wsum_secrets(secrets, bases):
    """
    Build expression representing a dot product of given secrets and bases.

    >>> from zksk.utils import make_generators
    >>> x, y = Secret(), Secret()
    >>> g, h = make_generators(2)
    >>> expr = wsum_secrets([x, y], [g, h])
    >>> expr.bases == (g, h)
    True
    >>> expr.secrets == (x, y)
    True

    Args:
        secrets: :py:class:`Secret` objects :math`x_i`
        bases: Elliptic curve points :math:`G_i`

    Returns:
        Expression: Expression that corresponds to :math:`x_0 G_0 + x_1 G_1 + ... + x_n G_n`
    """
    if len(secrets) != len(bases):
        raise ValueError("Should have as many secrets as bases.")

    result = secrets[0] * bases[0]
    for idx in range(len(bases) - 1):
        result = result + secrets[idx + 1] * bases[idx + 1]
    return result


def update_secret_values(secrets_dict):
    """
    Update values of secrets according to the given mapping.

    >>> x, y = Secret(), Secret()
    >>> secrets_dict = {x: 1, y: 2}
    >>> update_secret_values(secrets_dict)
    >>> x.value
    1
    >>> y.value
    2

    Args:
        secrets_dict: A mapping from :py:class:`Secret` objects to their expected values.
    """
    for k, v in secrets_dict.items():
        k.value = v



"""
Common exception classes.
"""


class IncompleteValuesError(Exception):
    """Cannot evaluate a proof as not all secret values are set."""


class InvalidExpression(Exception):
    pass


class InconsistentChallengeError(Exception):
    """Recomputed and global challenge values do not match."""


class StatementMismatch(Exception):
    """Proof statements mismatch, impossible to verify."""


class StatementSpecError(Exception):
    """Statement not fully specified."""


class InvalidSecretsError(Exception):
    """Secrets re-occur in an unsupported manner."""


class GroupMismatchError(Exception):
    """Generator groups mismatch."""


class VerificationError(Exception):
    """Error during verification."""


class ValidationError(Exception):
    """Error during validation."""



"""
Extended proofs are proofs that deal with internal precommitments.

This is a convenient building block that is useful for defining complex primitives.

Precommitments are parameters that are not known at the proof instantiation time
and are computed by the proving side. Therefore, they have to be sent explicitly
in a preliminary round for interactive protocols, and as an additional attribute
in non-interactive or simulated transcripts.

The protocol of an extended proof is as follows:

1. A proof statement (:py:class:`ExtendedProofStmt`) is instantiated and its
   attributes are set, but cannot run most methods.
2. Upon processing of the precommitment
   (:py:meth:`ExtendedProofProver.process_precommitment`), a separate and complete
   regular proof statement (:py:class:`zksk.base.ComposableProofStmt`) is constructed inside the
   extended proof (:py:attr`ExtendedProofStmt.constructed_stmt`)
3. :py:meth:`ExtendedProofStmt.get_bases`,
   :py:meth:`ExtendedProofStmt.get_secret_vars` delegate to the internal
   ``constructed_stmt``. The :py:class:`ExtendedProver` and
   :py:class:`ExtendedVerifier` work in the same way, embedding a
   ``constructed_prover`` (resp., ``constructed_verifier``).
"""

import abc

from zksk.base import Prover, Verifier
from zksk.composition import ComposableProofStmt
from zksk.exceptions import StatementSpecError
from zksk.utils.misc import get_default_attr


class ExtendedProofStmt(ComposableProofStmt, metaclass=abc.ABCMeta):
    """
    Proof that deals with precommitments.
    """

    @abc.abstractmethod
    def construct_stmt(self):
        """
        Build internal proof statement for this class

        This function must be overridden. The function should return a
        constructed proof statement. It can use the values that were computed
        by internal_precommit to do so.
        """
        pass

    def precommit(self):
        """
        Compute precommitments. Override if needed.

        Override this function to compute precommitments and set corresponding
        secrets that must be computed before the ZK proof itself can be
        constructed and proven.

        Returns:
            Precommitment
        """
        return []

    def simulate_precommit(self):
        """
        Simulate a precommitment. Override if needed.

        Override this method to enable using this proof in or-proofs.

        It should compute the same output as generated by precommit, but without relying on any
        secrets.
        """
        raise StatementSpecError(
            "Override simulate_precommit in order to " "use or-proofs and simulations"
        )

    def validate(self, precommitment, *args, **kwargs):
        """
        Validate proof's construction. Override if needed.
        """
        pass

    def full_validate(self, *args, **kwargs):
        self.validate(self.precommitment, *args, **kwargs)

    @property
    def constructed_stmt(self):
        return get_default_attr(self, "_constructed_stmt")

    @property
    def precommitment(self):
        return get_default_attr(self, "_precommitment")

    def get_secret_vars(self):
        return self.constructed_stmt.get_secret_vars()

    def get_bases(self):
        return self.constructed_stmt.get_bases()

    def get_prover(self, secrets_dict=None):
        """
        Get a prover object.

        Returns:
            Prover object if all secret values are known, None otherwise.
        """
        if secrets_dict is None:
            secrets_dict = {}

        for k, v in secrets_dict.items():
            k.value = v

        self.secret_values = {}
        self.secret_values.update(secrets_dict)

        return ExtendedProver(self, self.secret_values)

    def get_verifier_cls(self):
        return ExtendedVerifier

    def recompute_commitment(self, challenge, responses):
        return self.constructed_stmt.recompute_commitment(challenge, responses)

    def get_proof_id(self, secret_id_map=None):
        """
        Identifier for the proof statement.
        """
        if self.constructed_stmt is not None:
            proof_id = [
                self.__class__.__name__,
                self.precommitment,
                self.constructed_stmt.get_proof_id(secret_id_map),
            ]
        else:
            raise ValueError("Proof ID unknown before the proof is constructed.")
        return proof_id

    def full_construct_stmt(self, precommitment):
        self._precommitment = precommitment
        self._constructed_stmt = self.construct_stmt(precommitment)
        return self.constructed_stmt

    def prepare_simulate_proof(self):
        self._precommitment = self.simulate_precommit()
        self.full_construct_stmt(self.precommitment)

    def simulate_proof(self, responses_dict=None, challenge=None):
        """
        Simulate the proof.

        Args:
            responses_dict: Mapping from secrets to responses
            challenge: Challenge
        """
        tr = self._constructed_stmt.simulate_proof(
            challenge=challenge, responses_dict=responses_dict
        )
        tr.precommitment = self._precommitment
        return tr

    def _precommit(self):
        self._precommitment = self.precommit()
        return self._precommitment


class ExtendedProver(Prover):
    """
    Prover dealing with precommitments.

    This prover will create a constructed Prover object and delegate to its methods.
    """

    def internal_commit(self, randomizers_dict=None):
        """
        Trigger the internal prover commit.

        Transfers the randomizer_dict if passed. It might be used if the binding
        of the proof is set True.
        """
        if self.stmt.constructed_stmt is None:
            raise StatementSpecError(
                "You need to pre-commit before commiting. The proofs lack parameters otherwise."
            )
        return self.constructed_prover.internal_commit(randomizers_dict)

    def compute_response(self, challenge):
        """
        Wrap the response computation for the inner proof.
        """
        self.challenge = challenge
        self.constructed_prover.challenge = challenge
        self.response = self.constructed_prover.compute_response(challenge)
        return self.response

    def precommit(self):
        self.precommitment = self.stmt._precommit()
        self.process_precommitment()
        try:
            for p in self.constructed_prover.subs:
                p.precommit()
        except AttributeError:
            pass
        return self.precommitment

    def process_precommitment(self):
        """
        Trigger the inner-proof construction and extract a prover given the secrets.
        """
        self.stmt.full_construct_stmt(self.precommitment)
        self.constructed_prover = self.stmt._constructed_stmt.get_prover(
            self.secret_values
        )


class ExtendedVerifier(Verifier):
    """
    Verifier that deals with precommitments.
    """

    def process_precommitment(self, precommitment):
        """
        Receive the precommitment and trigger the inner-verifier construction.
        """
        self.precommitment = precommitment
        self.stmt.full_construct_stmt(precommitment)
        self.constructed_verifier = self.stmt.constructed_stmt.get_verifier()

    def send_challenge(self, com):
        """
        Transfer the commitment to the inner proof, and compute the challenge.
        """

        statement, self.commitment = com
        self.stmt.check_statement(statement)
        self.challenge = self.constructed_verifier.send_challenge(
            self.commitment, ignore_statement_hash_checks=True
        )
        return self.challenge

    def check_responses_consistency(self, responses, responses_dict):
        """
        Wrap the consistency check of the innter proof.
        """
        return self.constructed_verifier.check_responses_consistency(
            responses, responses_dict
        )



__version__ = "0.0.31"
__title__ = "zksk"
__author__ = "Wouter Lueks, Bogdan Kulynych, Jules Fasquelle, Simon Le Bail-Collet"
__email__ = "wouter.lueks@epfl.ch"
__url__ = "https://zksk.readthedocs.io"
__license__ = "MIT"
__description__ = "Zero-Knowledge Swiss Knife: Python library for prototyping composable zero-knowledge proofs."
__copyright__ = "2021, Wouter Lueks, Bogdan Kulynych (EPFL SPRING Lab)"


from zksk.expr import Secret
from zksk.primitives.dlrep import DLRep
from zksk.utils import make_generators
